/*
 * This file is part of htmltolatex (h2l) a HTML to Latex converter
 * Copyright (C) 2001 Neil Radisch (nradisch@panix.com)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file copying.txt.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */
#include "CH2L.h"
#include "util.h"
#include "safe_ostrstream.h"
#include "Magick++.h"	// imageMagick
#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <fstream>

using namespace std;
using namespace Magick;

#define ATTRIBUTENESTLIMIT	1000
#define ALLOW_CELL_ALIGNMENT 1


static const string	IA_FONT_SIZE	="@@@font_size";
static const string IA_FONT_NAME	="@@@font_name";
static const string IA_FONT_SERIES	="@@@font_series";
static const string IA_FONT_SHAPE	="@@@font_shape";
static const string IA_FONT_COLOR	="@@@font_color";

static const string IA_PAR_ALIGN	="@@@par_align";
static const string IA_IMG_ALIGN	="@@@img_align";
static const string IA_TAB_ALIGN	="@@@tab_align";
static const string IA_SPCR_ALIGN	="@@@spcr_align";
static const string IA_MIN_CELLWIDTH	= "@@@tab_minwidth";

static const string IA_JUNKTAG		="@@@@@@@@";

static const string IA_TABLE_DIM_INSTANCE	="@@@table_dim_instance";

/******************************************
Custom attribute and Attribute name conversions

align		IA_SPCR_ALIGN		spacer align
align		IA_TAB_ALIGN		table aligns
align 		IA_PAR_ALIGN		paragraph aligns
align		IA_IMG_ALIGN		image aligns
			IA_FONT_NAME		font name - latex font name
			IA_FONT_SERIES		1 = bold
			IA_FONT_SHAPE		1 = italic
								2 = slanted
								4 = small caps
size		IA_FONT_SIZE		font size in points
*******************************************/
static const string DEFAULT_DOCUMENT_ALIGN="justify";
static const string	DEFAULT_HEADING_ALIGN="justify";
static const string DEFAULT_PARAGRAPH_ALIGN="justify";
static const string DEFAULT_RULE_ALIGN="center";
static const string DEFAULT_FONT_COLOR="black";
static const size_t	MAX_TABLE_COLUMNS=10;

#define DEFAULT_FONT_SIZE	((double)(10.0))
static const string DEFAULT_DOCUMENT_FONT="ptm";

static string gsCharTable[256] = 
{
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
"~\\newline ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
" ",
"\\ ",
"!",
"$\"$", // dble quote
"{\\#}",
"{\\$}",
"{\\%}",
"{\\&}",
"'",
"(",
")",
"*",
"+",
",",
"-",
".",
"/",
"0",
"1",
"2",
"3",
"4",
"5",
"6",
"7",
"8",
"9",
":",
";",
"$<$",
"=",
"$>$",
"?",
"@",
"A",
"B",
"C",
"D",
"E",
"F",
"G",
"H",
"I",
"J",
"K",
"L",
"M",
"N",
"O",
"P",
"Q",
"R",
"S",
"T",
"U",
"V",
"W",
"X",
"Y",
"Z",
"[",
"$\\backslash$",
"]",
"\\^{ }",
"{\\_}",
"`",
"a",
"b",
"c",
"d",
"e",
"f",
"g",
"h",
"i",
"j",
"k",
"l",
"m",
"n",
"o",
"p",
"q",
"r",
"s",
"t",
"u",
"v",
"w",
"x",
"y",
"z",
"$\\{$",
"$|$",
"$\\}$",
"$\\sim$",		// tilde
" ",
" ",
" ",
",",			// ????? asci 130
"{\\textflorin}",// florin
"''",			// right dble quote
"{\\dots}",		// elipses
"{\\dag}",		// dagger
"{\\ddag}",		// double dagger
"\\^{ }",		// circumflex
" ",			// permil - dont have
"{\\_}",		// ????? asci 138
"$<$",			// less than (again)
"{\\OE}",		// OE ligature
" ",
" ",
" ",
" ",
"`",			// left single quote
"'",			// right single quote
"``",			// left dble quote
"''",			// right dble quote
"$\\bullet$",	// bullet
"--",			// en dash
"---",			// em dash
"$\\sim$",		// tilde (again)
"{\\texttrademark}",		// trademark
"{\\_}",		// ????? asci 154
"$>$",			// greater than (again)
"{\\oe}",		// oe ligature
" ",
" ",
"\\\"{Y}",		// Y with umlat
"~",			// non breaking space
"!`",			// inverted exclamation
"{\\textcent}",	// cents symbol
"{\\textsterling}",	// pound sterling
"{\\textcurrency}",	// general curreny sign
"{\\textyen}",	// Yen symbol
"{\\textbrokenbar}",// broken bar
"{\\S}",		// section symbol
"\\\"{ }",		// lone umlat
"{\\copyright}",	// copyright symbol
"$^{\\hbox{\\rlap a\\_}}$",	// feminine ordinal
"``",			// left dbl quote
"$\\neg$",		// not sign
"-",			// hyphen
"{\\textregistered}",	// circle r
"\\={ }",		// macron accent (bar)
"{\\textdegree}",// degree
"$\\pm$",		// plus or minus
"$^{2}$",		// superscript 2
"$^{3}$",		// superscript 3
"\\'{ }",		// acute accent
"$\\mu$",		// micro symbol
"{\\P}",		// paragraph symbol
"$\\cdot$",		// middle dot
"\\c{ }",		// cedilla
"$^{1}$",		// superscript 1
"$^{\\hbox{\\rlap o\\_}}$",	// feminine ordinal
"''",			// right angle quote
"$1\\over4$",	// 1/4
"$1\\over2$",	// 1/2
"$3\\over4$",	// 3/4
"?`",			// inverted question mark
"{\\`A}",		// A grave
"{\\'A}",		// A acute
"{\\^A}",		// A circumflex
"{\\~A}",		// A tilde
"{\\\"A}",		// A umlat
"{\\.A}",		// A dot
"{\\AE}",		// AE ligature
"{\\c C}",		// C cedilla
"{\\`E}",		// E grave
"{\\'E}",		// E acute
"{\\^E}",		// E circumflex
"{\\\"E}",		// E umlat
"{\\`I}",		// I grave
"{\\'I}",		// I acute
"{\\^I}",		// I circumflex
"{\\\"I}",		// I umlat
"{\\hbox{\\rlap D-} }",	// eth
"{\\~N}",		// N tilde
"{\\`O}",		// O grave
"{\\'O}",		// O acute
"{\\^O}",		// O circumflex
"{\\~O}",		// O tilde
"{\\\"O}",		// O umlat
"$\\times$",	// multiply
"{\\O}",		// O slash
"{\\`U}",		// U grave
"{\\'U}",		// U acute
"{\\^U}",		// U circumflex
"{\\\"U}",		// U umlat
"{\\'Y}",		// Y acute
"b",			// capital THORN ???/
"{\\ss}",		// german es-zet ligature
"{\\`a}",		// a grave
"{\\'a}",		// a acute
"{\\^a}",		// a circumflex
"{\\~a}",		// a tilde
"{\\\"a}",		// a umlat
"{\\.a}",		// a dot
"{\\ae}",		// ae ligature
"{\\c c}",		// c cedilla
"{\\`e}",		// e grave
"{\\'e}",		// e acute
"{\\^e}",		// e circumflex
"{\\\"e}",		// e umlat
"{\\`i}",		// i grave
"{\\'i}",		// i acute
"{\\^i}",		// i circumflex
"{\\\"i}",		// i umlat
"d",			// small eth (dont have yet)
"{\\~n}",		// n tilde
"{\\`o}",		// o grave
"{\\'o}",		// o acute
"{\\^o}",		// o circumflex
"{\\~o}",		// o tilde
"{\\\"o}",		// o umlat
"$\\div$",		// division
"{\\o}",		// o slash
"{\\`u}",		// u grave
"{\\'u}",		// u acute
"{\\^u}",		// u circumflex
"{\\\"u}",		// u umlat
"{\\'y}",		// y acute
"b",			// small THORN ???/
"{\\\"y}"		// y umlat
};

CH2L::TableState::LastTableTag_t CH2L::TableState::mLastTableTag;


CH2L::CH2L(): mColorWhiteTables(true),mVerbatimMode(false),mWaitingForListItem(false),mWaitingForTableElement(false),
				mAllowHyperlinks(true),
				mAllowFileFetch(false),mProcessImages(true),mProcessTables(true),mProcessStyles(true),
				mPaperWidth("8.5in"),
				mPaperHeight("11in"),
				mHorizontalMargin("0.5in"),
				mVerticalMargin("0.5in"),
				mTempPath("/var/tmp/"),
				mOrderedListDepth(0),mSpidering(false),mSpiderTagCounter(0),mAlignmentString(DEFAULT_DOCUMENT_ALIGN),
				mInlineTables(false),mTableDimInstanceCtr(0)
{ 
	mSingleTags.insert(mSingleTags.begin(),"optgroup");
	mSingleTags.insert(mSingleTags.begin(),"/optgroup");
	mSingleTags.insert(mSingleTags.begin(),"embed");
	mSingleTags.insert(mSingleTags.begin(),"/embed");
	mSingleTags.insert(mSingleTags.begin(),"option");
	mSingleTags.insert(mSingleTags.begin(),"/option");
	mSingleTags.insert(mSingleTags.begin(),"param");
	mSingleTags.insert(mSingleTags.begin(),"/param");
	mSingleTags.insert(mSingleTags.begin(),"frame");
	mSingleTags.insert(mSingleTags.begin(),"/frame");
	mSingleTags.insert(mSingleTags.begin(),"col");
	mSingleTags.insert(mSingleTags.begin(),"/col");
	mSingleTags.insert(mSingleTags.begin(),"colgroup");
	mSingleTags.insert(mSingleTags.begin(),"/colgroup");
	mSingleTags.insert(mSingleTags.begin(),"tbody");
	mSingleTags.insert(mSingleTags.begin(),"/tbody");
	mSingleTags.insert(mSingleTags.begin(),"tfoot");
	mSingleTags.insert(mSingleTags.begin(),"/tfoot");
	mSingleTags.insert(mSingleTags.begin(),"thread");
	mSingleTags.insert(mSingleTags.begin(),"/thread");
	mSingleTags.insert(mSingleTags.begin(),"nextid");
	mSingleTags.insert(mSingleTags.begin(),"base");
	mSingleTags.insert(mSingleTags.begin(),"meta");
	mSingleTags.insert(mSingleTags.begin(),"area");
	mSingleTags.insert(mSingleTags.begin(),"isindex");
	mSingleTags.insert(mSingleTags.begin(),"basefont");
	mSingleTags.insert(mSingleTags.begin(),"/basefont");
	mSingleTags.insert(mSingleTags.begin(),"input");
	mSingleTags.insert(mSingleTags.begin(),"spacer");
	mSingleTags.insert(mSingleTags.begin(),"bgsound");
	mSingleTags.insert(mSingleTags.begin(),"img");
	mSingleTags.insert(mSingleTags.begin(),"dd");
	mSingleTags.insert(mSingleTags.begin(),"/dd");
	mSingleTags.insert(mSingleTags.begin(),"dt");
	mSingleTags.insert(mSingleTags.begin(),"/dt");
	mSingleTags.insert(mSingleTags.begin(),"/li");
	mSingleTags.insert(mSingleTags.begin(),"li");
	mSingleTags.insert(mSingleTags.begin(),"hr");
	mSingleTags.insert(mSingleTags.begin(),"br");
	mSingleTags.insert(mSingleTags.begin(),"p");
	mSingleTags.insert(mSingleTags.begin(),"/p");
	mSingleTags.insert(mSingleTags.begin(),"form");
	mSingleTags.insert(mSingleTags.begin(),"/form");
	mSingleTags.insert(mSingleTags.begin(),"link");
	mSingleTags.insert(mSingleTags.begin(),"tr");
	mSingleTags.insert(mSingleTags.begin(),"td");
	mSingleTags.insert(mSingleTags.begin(),"th");
	mSingleTags.insert(mSingleTags.begin(),"/tr");
	mSingleTags.insert(mSingleTags.begin(),"/td");
	mSingleTags.insert(mSingleTags.begin(),"/th");
	mSingleTags.insert(mSingleTags.begin(),"b");
	mSingleTags.insert(mSingleTags.begin(),"/b");
	mSingleTags.insert(mSingleTags.begin(),"strong");
	mSingleTags.insert(mSingleTags.begin(),"/strong");
	mSingleTags.insert(mSingleTags.begin(),"blink");
	mSingleTags.insert(mSingleTags.begin(),"/blink");
	mSingleTags.insert(mSingleTags.begin(),"i");
	mSingleTags.insert(mSingleTags.begin(),"/i");
	mSingleTags.insert(mSingleTags.begin(),"abbr");
	mSingleTags.insert(mSingleTags.begin(),"/abbr");
	mSingleTags.insert(mSingleTags.begin(),"cite");
	mSingleTags.insert(mSingleTags.begin(),"/cite");
	mSingleTags.insert(mSingleTags.begin(),"em");
	mSingleTags.insert(mSingleTags.begin(),"/em");
	mSingleTags.insert(mSingleTags.begin(),"address");
	mSingleTags.insert(mSingleTags.begin(),"/address");
	mSingleTags.insert(mSingleTags.begin(),"acronym");
	mSingleTags.insert(mSingleTags.begin(),"/acronym");
	mSingleTags.insert(mSingleTags.begin(),"code");
	mSingleTags.insert(mSingleTags.begin(),"/code");
	mSingleTags.insert(mSingleTags.begin(),"var");
	mSingleTags.insert(mSingleTags.begin(),"/var");
	mSingleTags.insert(mSingleTags.begin(),"kbd");
	mSingleTags.insert(mSingleTags.begin(),"/kbd");
	mSingleTags.insert(mSingleTags.begin(),"tt");
	mSingleTags.insert(mSingleTags.begin(),"/tt");
	mSingleTags.insert(mSingleTags.begin(),"ins");
	mSingleTags.insert(mSingleTags.begin(),"/ins");
	mSingleTags.insert(mSingleTags.begin(),"del");
	mSingleTags.insert(mSingleTags.begin(),"/del");
	mSingleTags.insert(mSingleTags.begin(),"strike");
	mSingleTags.insert(mSingleTags.begin(),"/strike");
	mSingleTags.insert(mSingleTags.begin(),"s");
	mSingleTags.insert(mSingleTags.begin(),"/s");
	mSingleTags.insert(mSingleTags.begin(),"u");
	mSingleTags.insert(mSingleTags.begin(),"/u");
	mSingleTags.insert(mSingleTags.begin(),"nobr");
	mSingleTags.insert(mSingleTags.begin(),"/nobr");
	mSingleTags.insert(mSingleTags.begin(),"big");
	mSingleTags.insert(mSingleTags.begin(),"/big");
	mSingleTags.insert(mSingleTags.begin(),"small");
	mSingleTags.insert(mSingleTags.begin(),"/small");
	mSingleTags.insert(mSingleTags.begin(),"font");
	mSingleTags.insert(mSingleTags.begin(),"/font");
	mSingleTags.insert(mSingleTags.begin(),"span");
	mSingleTags.insert(mSingleTags.begin(),"/span");
	mSingleTags.insert(mSingleTags.begin(),"div");
	mSingleTags.insert(mSingleTags.begin(),"/div");
	mSingleTags.insert(mSingleTags.begin(),"dfn");
	mSingleTags.insert(mSingleTags.begin(),"/dfn");
	mSingleTags.insert(mSingleTags.begin(),"samp");
	mSingleTags.insert(mSingleTags.begin(),"/samp");
	mSingleTags.insert(mSingleTags.begin(),"a");
	mSingleTags.insert(mSingleTags.begin(),"/a");
	mSingleTags.insert(mSingleTags.begin(),"lh");
	mSingleTags.insert(mSingleTags.begin(),"/lh");
	mSingleTags.insert(mSingleTags.begin(),"center");
	mSingleTags.insert(mSingleTags.begin(),"/center");

	mFontNameFaceMap["avantgarde"] 	= "pag";
	mFontNameFaceMap["avant garde"] = "pag";
	mFontNameFaceMap["bookman"] 	= "pbk";
	mFontNameFaceMap["courier"] 	= "pcr";
	mFontNameFaceMap["courier new"] = "pcr";
	mFontNameFaceMap["fixed"] 		= "pcr";
	mFontNameFaceMap["helvetica"] 	= "phv";
	mFontNameFaceMap["arial"] 		= "phv";
	mFontNameFaceMap["verdana"] 	= "phv";
	mFontNameFaceMap["sansserif"] 	= "phv";
	mFontNameFaceMap["sans serif"] 	= "phv";
	mFontNameFaceMap["newcenturyschoolbook"] = "pnc";
	mFontNameFaceMap["new century schoolbook"] = "pnc";
	mFontNameFaceMap["palatino"] 	= "ppl";
	mFontNameFaceMap["times"] 		= "ptm";
	mFontNameFaceMap["roman"] 		= "ptm";
	mFontNameFaceMap["times roman"] = "ptm";
	mFontNameFaceMap["times new roman"] = "ptm";
	mFontNameFaceMap["zapfchancery"] = "pzc";

	mRelativeFontList.push_back(5);
	mRelativeFontList.push_back(7);
	mRelativeFontList.push_back(8);
	mRelativeFontList.push_back(9);
	mRelativeFontList.push_back(10);
	mRelativeFontList.push_back(12);
	mRelativeFontList.push_back(14.4);
	mRelativeFontList.push_back(17.28);
	mRelativeFontList.push_back(20.736);
	mRelativeFontList.push_back(24.8832);

	mParagraphResetTags.insert(mParagraphResetTags.begin(),"p");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"/p");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"hr");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"div");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"span");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"h6");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"h5");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"h4");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"h3");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"h2");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"h1");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"center");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"table");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"tr");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"td");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"/td");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"th");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"form");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"ol");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"ul");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"li");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"lh");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"dl");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"dt");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"dd");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"pre");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"listing");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"xmp");
	mParagraphResetTags.insert(mParagraphResetTags.begin(),"plaintext");


	mMakeMethodMap["hr"] = &CH2L::make_horizontal_rule;
	mMakeMethodMap["br"] = &CH2L::make_linebreak;
	mMakeMethodMap["p"] = &CH2L::make_paragraph_start;
	mMakeMethodMap["/p"] = &CH2L::make_paragraph_end;
	mMakeMethodMap["basefont"] 	= &CH2L::make_basefont_start;
	mMakeMethodMap["/basefont"] = &CH2L::make_basefont_end;
	mMakeMethodMap["img"]		= &CH2L::make_image;
	mMakeMethodMap["spacer"]	= &CH2L::make_spacer;
	mMakeMethodMap["li"]		= &CH2L::make_li_start;
	mMakeMethodMap["lh"]		= &CH2L::make_lh_start;
	mMakeMethodMap["dt"]		= &CH2L::make_dt_start;
	mMakeMethodMap["dd"]		= &CH2L::make_dd_start;
	mMakeMethodMap["tr"]		= &CH2L::make_tr_start;
	mMakeMethodMap["td"]		= &CH2L::make_td_start;
	mMakeMethodMap["th"]		= &CH2L::make_th_start;
	mMakeMethodMap["/tr"]		= &CH2L::make_tr_end;
	mMakeMethodMap["input"]		= &CH2L::make_input;
	mMakeMethodMap["option"]	= &CH2L::make_option_start;
	mMakeMethodMap["/option"]	= &CH2L::make_option_end;
	mMakeMethodMap["b"] 		= &CH2L::make_bold_start;
	mMakeMethodMap["/b"]		= &CH2L::make_bold_end;
	mMakeMethodMap["strong"] 	= &CH2L::make_strong_start;
	mMakeMethodMap["/strong"]	= &CH2L::make_strong_end;
	mMakeMethodMap["blink"] 	= &CH2L::make_blink_start;
	mMakeMethodMap["/blink"]	= &CH2L::make_blink_end;
	mMakeMethodMap["i"] 		= &CH2L::make_italic_start;
	mMakeMethodMap["/i"]		= &CH2L::make_italic_end;
	mMakeMethodMap["abbr"] 		= &CH2L::make_abbr_start;
	mMakeMethodMap["/abbr"]		= &CH2L::make_abbr_end;
	mMakeMethodMap["cite"] 		= &CH2L::make_cite_start;
	mMakeMethodMap["/cite"]		= &CH2L::make_cite_end;
	mMakeMethodMap["em"] 		= &CH2L::make_em_start;
	mMakeMethodMap["/em"]		= &CH2L::make_em_end;
	mMakeMethodMap["address"] 	= &CH2L::make_address_start;
	mMakeMethodMap["/address"]	= &CH2L::make_address_end;
	mMakeMethodMap["acronym"] 	= &CH2L::make_acronym_start;
	mMakeMethodMap["/acronym"]	= &CH2L::make_acronym_end;
	mMakeMethodMap["code"] 		= &CH2L::make_code_start;
	mMakeMethodMap["/code"]		= &CH2L::make_code_end;
	mMakeMethodMap["var"] 		= &CH2L::make_var_start;
	mMakeMethodMap["/var"]		= &CH2L::make_var_end;
	mMakeMethodMap["kbd"] 		= &CH2L::make_kbd_start;
	mMakeMethodMap["/kbd"]		= &CH2L::make_kbd_end;
	mMakeMethodMap["tt"] 		= &CH2L::make_tt_start;
	mMakeMethodMap["/tt"]		= &CH2L::make_tt_end;
	mMakeMethodMap["ins"] 		= &CH2L::make_ins_start;
	mMakeMethodMap["/ins"]		= &CH2L::make_ins_end;
	mMakeMethodMap["del"] 		= &CH2L::make_del_start;
	mMakeMethodMap["/del"]		= &CH2L::make_del_end;
	mMakeMethodMap["strike"] 	= &CH2L::make_strike_start;
	mMakeMethodMap["/strike"]	= &CH2L::make_strike_end;
	mMakeMethodMap["s"] 		= &CH2L::make_s_start;
	mMakeMethodMap["/s"]		= &CH2L::make_s_end;
	mMakeMethodMap["u"] 		= &CH2L::make_u_start;
	mMakeMethodMap["/u"]		= &CH2L::make_u_end;
	mMakeMethodMap["nobr"] 		= &CH2L::make_nobr_start;
	mMakeMethodMap["/nobr"]		= &CH2L::make_nobr_end;
	mMakeMethodMap["big"] 		= &CH2L::make_big_start;
	mMakeMethodMap["/big"]		= &CH2L::make_big_end;
	mMakeMethodMap["small"] 	= &CH2L::make_small_start;
	mMakeMethodMap["/small"]	= &CH2L::make_small_end;
	mMakeMethodMap["font"] 		= &CH2L::make_font_start;
	mMakeMethodMap["/font"]		= &CH2L::make_font_end;
	mMakeMethodMap["span"] 		= &CH2L::make_span_start;
	mMakeMethodMap["/span"]		= &CH2L::make_span_end;
	mMakeMethodMap["div"] 		= &CH2L::make_div_start;
	mMakeMethodMap["/div"]		= &CH2L::make_div_end;
	mMakeMethodMap["dfn"] 		= &CH2L::make_dfn_start;
	mMakeMethodMap["/dfn"]		= &CH2L::make_dfn_end;
	mMakeMethodMap["samp"] 		= &CH2L::make_samp_start;
	mMakeMethodMap["/samp"]		= &CH2L::make_samp_end;
	mMakeMethodMap["a"] 		= &CH2L::make_a_start;
	mMakeMethodMap["/a"]		= &CH2L::make_a_end;
	mMakeMethodMap["center"] 	= &CH2L::make_center_start;
	mMakeMethodMap["/center"]	= &CH2L::make_center_end;

	mStartMethodMap["body"] = &CH2L::start_body;
	mEndMethodMap["body"] 	= &CH2L::end_body;
	mStartMethodMap["h1"] 	= &CH2L::start_h1;
	mEndMethodMap["h1"] 	= &CH2L::end_h1;
	mStartMethodMap["h2"] 	= &CH2L::start_h2;
	mEndMethodMap["h2"] 	= &CH2L::end_h2;
	mStartMethodMap["h3"] 	= &CH2L::start_h3;
	mEndMethodMap["h3"] 	= &CH2L::end_h3;
	mStartMethodMap["h4"] 	= &CH2L::start_h4;
	mEndMethodMap["h4"] 	= &CH2L::end_h4;
	mStartMethodMap["h5"] 	= &CH2L::start_h5;
	mEndMethodMap["h5"] 	= &CH2L::end_h5;
	mStartMethodMap["h6"] 	= &CH2L::start_h6;
	mEndMethodMap["h6"] 	= &CH2L::end_h6;
	mStartMethodMap["sub"] 		= &CH2L::start_sub;
	mEndMethodMap["sub"]		= &CH2L::end_sub;
	mStartMethodMap["sup"] 		= &CH2L::start_sup;
	mEndMethodMap["sup"]		= &CH2L::end_sup;
	mStartMethodMap["pre"] 		= &CH2L::start_pre;
	mEndMethodMap["pre"]		= &CH2L::end_pre;
	mStartMethodMap["listing"] 	= &CH2L::start_listing;
	mEndMethodMap["listing"]	= &CH2L::end_listing;
	mStartMethodMap["xmp"] 		= &CH2L::start_xmp;
	mEndMethodMap["xmp"]		= &CH2L::end_xmp;
	mStartMethodMap["plaintext"]= &CH2L::start_plaintext;
	mEndMethodMap["plaintext"]	= &CH2L::end_plaintext;
	mStartMethodMap["blockquote"]= &CH2L::start_blockquote;
	mEndMethodMap["blockquote"]	= &CH2L::end_blockquote;
	mStartMethodMap["q"]		= &CH2L::start_q;
	mEndMethodMap["q"]			= &CH2L::end_q;
	mStartMethodMap["marquee"]	= &CH2L::start_marquee;
	mEndMethodMap["marquee"]	= &CH2L::end_marquee;
	mStartMethodMap["multicol"]	= &CH2L::start_multicol;
	mEndMethodMap["multicol"]	= &CH2L::end_multicol;
	mStartMethodMap["ol"]		= &CH2L::start_ol;
	mEndMethodMap["ol"]			= &CH2L::end_ol;
	mStartMethodMap["ul"]		= &CH2L::start_ul;
	mEndMethodMap["ul"]			= &CH2L::end_ul;
	mStartMethodMap["dir"]		= &CH2L::start_dir;
	mEndMethodMap["dir"]		= &CH2L::end_dir;
	mStartMethodMap["menu"]		= &CH2L::start_menu;
	mEndMethodMap["menu"]		= &CH2L::end_menu;
	mStartMethodMap["dl"]		= &CH2L::start_dl;
	mEndMethodMap["dl"]			= &CH2L::end_dl;
	mStartMethodMap["table"]	= &CH2L::start_table;
	mEndMethodMap["table"]		= &CH2L::end_table;
	mStartMethodMap["button"]	= &CH2L::start_button;
	mEndMethodMap["button"]		= &CH2L::end_button;
	mStartMethodMap["textarea"]	= &CH2L::start_textarea;
	mEndMethodMap["textarea"]	= &CH2L::end_textarea;
	mStartMethodMap["select"]	= &CH2L::start_select;
	mEndMethodMap["select"]		= &CH2L::end_select;

	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"img");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"body");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"br");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"hr");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"p");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"/p");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"center");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"/center");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"h1");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"h2");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"h3");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"h4");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"h5");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"h6");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"address");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"blockquote");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"ol");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"/ol");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"ul");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"/ul");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"li");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"/li");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"lh");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"dd");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"dt");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"dl");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"/dd");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"/dt");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"/dl");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"table");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"/table");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"tr");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"/tr");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"td");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"/td");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"th");
	mIgnoreSpaceTags.insert(mIgnoreSpaceTags.begin(),"/th");

	reset_attribute_stack();
}

bool CH2L::tag_ignores_spaces(const string &tag) const
{
	return (mIgnoreSpaceTags.find(tag) != mIgnoreSpaceTags.end());
}

void CH2L::tag_make(const string &t)
{
	MakeMethodMap_t::const_iterator i = mMakeMethodMap.find(t);
	if (i != mMakeMethodMap.end())
		{
		makemethod_t m = (*i).second;
		(this->*m)();
		}
	else
		{
		dump_text();
		_pop_attribute_layer();
		}
}

void CH2L::tag_start(const string &t)
{
	StartMethodMap_t::const_iterator i = mStartMethodMap.find(t);
	if (i != mStartMethodMap.end())
		{
		startmethod_t m = (*i).second;
		(this->*m)();
		}
	else
		{
		dump_text();
		}
}

void CH2L::tag_end(const string &t)
{
	EndMethodMap_t::const_iterator i = mEndMethodMap.find(t);
	if (i != mEndMethodMap.end())
		{
		endmethod_t m = (*i).second;
		(this->*m)();
		}
	else
		{
		// just pop the attributes
		dump_text();
		_pop_attribute_layer();
		}
}

void CH2L::_html_color_to_latex_info(const string &sin,
								bool &rgb,
								bool &named,
								string &spec) const
{
	spec=""; rgb = false; named = false;

	string s = stripstring(sin);
	if (s == "") return;

	if ((s[0] == '#'))// && (s.size() == 7))
		{
		double r,g,b;
		htmlrgbtotexrgb(s,r,g,b);
		rgb = true;
		safe_ostrstream o;
		o << r << "," << g << "," << b;
		spec = o.str();
		}
	else
		{
		ColorMap_t::const_iterator i = mLegalColorNames.find(s);
		if (i != mLegalColorNames.end())
			{
			named = true;
			spec = s;
			}
		else
			{
			// see if we really have hex digits, but without the #
			if (s.size() == 6 && allhexdigits(s))
				{
				s = string("#") + s;
				double r,g,b;
				htmlrgbtotexrgb(s,r,g,b);
				rgb = true;
				safe_ostrstream o;
				o << r << "," << g << "," << b;
				spec = o.str();
				}
			}
		}
}

string CH2L::_html_color_to_latex_color(const string &sin) const
{
	string result;
	bool rgb;
	bool named;

	safe_ostrstream o;

	_html_color_to_latex_info(sin,rgb,named,result);
	if (rgb || named)
		{
		if (rgb)
			o << "[rgb]";
		o << "{" << result << "}";
		}
	return o.str();
}

void CH2L::_set_hyperlink_color(const string &s) const
{
	string result;
	bool rgb;
	bool named;	

	_html_color_to_latex_info(s,rgb,named,result);
	if (rgb || named)
		{
		if (rgb)
			{
			cout << "\\definecolor{hyperlinkcolor}{rgb}{" << result << "}" << endl;
			cout << "\\hypersetup{urlcolor=hyperlinkcolor,anchorcolor=hyperlinkcolor,linkcolor=hyperlinkcolor}" << endl;
			}
		else if (named)
			{
			ColorMap_t::const_iterator i = mLegalColorNames.find(result);
			if (i != mLegalColorNames.end())
				{
				const latex_rgb &lrgb = (*i).second;
				cout << "\\definecolor{hyperlinkcolor}{rgb}{" << lrgb.r << "," << lrgb.g << "," << lrgb.b << "}" << endl;
				cout << "\\hypersetup{urlcolor=hyperlinkcolor,anchorcolor=hyperlinkcolor,linkcolor=hyperlinkcolor}" << endl;
				}
			}
		
		}
}

void CH2L::generate_latex_prefix()
{
#if 0
	if (!mSpidering)
		{
		if (mTitle != "")
			cout << "\\title{" << mTitle << "}" << endl;
		}
#endif

	cout << "{"; // color bracket

	if (mTitle != "" && (!mSpidering))
		{
#if 0
		cout << "\\maketitle" << endl;
		mTitle = "";
#endif
		cout << "\\hypersetup{pdftitle={" << mTitle << "}}" << endl;
		}

	if (mCurrentSpiderTag != "")
		cout << "\\hypertarget{" << mCurrentSpiderTag << "}{}";

	cout << "\\" << DEFAULT_DOCUMENT_ALIGN << endl;
}


void CH2L::latex_preheader()
{
	if (mSpidering)
		{
		cout << endl << "}\\par\\pagebreak[4]" << endl;
		}
	if (!mSpidering)
		{
		cout << "\\documentclass[letterpaper]{article}" << endl;
		cout << "\\usepackage{geometry}" << endl;
		cout << "\\geometry{dvips,nohead," 
				<< "paperheight=" << mPaperHeight << ","
				<< "paperwidth=" << mPaperWidth << ","
				<< "hmargin=" << mHorizontalMargin << ","
				<< "vmargin=" << mVerticalMargin 
				<< "}" << endl;
		cout << "\\input{" << mPathToDataFile << "nmr}" << endl;
		if (!mInlineTables) 
			cout << "\\input{" << mTableDimensionsFileName << "}" << endl;
		cout << "\\usepackage[dvips," << (mURLServerRoot != "" ? (string("baseurl={") + mURLServerRoot + "},") : "") << "colorlinks=true]{hyperref}" << endl;

		cout << "\\begin{document}" << endl;
	
		///////////////////////////////////////////////////
		// load all standard X11 colors in from rgb.txt
		// search for different rgb.txt files
		vector<string> paths;
		paths.push_back("/usr/X11R6/lib/X11/rgb.txt");
		paths.push_back("/usr/lib/X11/rgb.txt");
		paths.push_back("rgb.txt");

		// these should be built in
		mLegalColorNames["black"] 	= latex_rgb(0,0,0);
		mLegalColorNames["white"] 	= latex_rgb(1,1,1);
		mLegalColorNames["red"] 	= latex_rgb(1,0,0);
		mLegalColorNames["green"] 	= latex_rgb(0,1,0);
		mLegalColorNames["blue"]	= latex_rgb(0,0,1);
		mLegalColorNames["cyan"] 	= latex_rgb(0,1,1);
		mLegalColorNames["magenta"] = latex_rgb(1,0,1);
		mLegalColorNames["yellow"] 	= latex_rgb(1,1,0);

		ifstream f;
		bool got_rgb_file = false;
		for(vector<string>::const_iterator i = paths.begin(); i != paths.end(); ++i)
			{
			f.open((*i).c_str());
			if (f)
				{
				got_rgb_file = true;
				break;
				}
			}
		if (got_rgb_file)
			{
			int ri,gi,bi;
			double r,g,b;
			char buf[256];

			while (f)
				{
				char c;
				f.get(c);
				if (!f) break;
				if (isdigit(c))
					{
					f.putback(c);
					f >> ri >> gi >> bi;
					r = ri; g = gi; b = bi;
					r /= 255.0;
					g /= 255.0;
					b /= 255.0;
					if (f)
						{
						f.getline(buf,255,'\n');
						string s = stripstring(string(buf));
						if (s != "")
							{
							mLegalColorNames[s] = latex_rgb(r,g,b);
							cout << "\\definecolor{" << s << "}{rgb}{" << r << "," << g << "," << b << "}" << endl;
							}
						}
					}
				else if (c != '\n')
					{
					char ln[512];
					f.getline(ln,511,'\n');
					}
				}
			f.close();
			}
		///////////////////////////////////////////
		_set_font_all();
		}

	generate_latex_prefix();
}

void CH2L::start_body()
{
#if 0
	if (mTitle != "" && (!mSpidering))
		{
		cout << "\\title{" << mTitle << "}" << endl;
		cout << "\\maketitle" << endl;
		mTitle = "";
		}
#endif

	Attribute a;
	if (_get_attribute("bgcolor",a))
		{
		string c = _html_color_to_latex_color(a.getStringVal());
		if (c != "")
			cout << "\\pagecolor" << c << endl;
		}
	if (_get_attribute("text",a))
		{
		_alter_attribute_name("text",IA_FONT_COLOR);
		_adjust_color(_html_color_to_latex_color(a.getStringVal()),true);
		}
	if (_get_attribute("link",a))
		{
		_set_hyperlink_color(a.getStringVal());
		}

	_change_alignment(DEFAULT_DOCUMENT_ALIGN,true);
}

void CH2L::end_body()
{
	dump_text();

	_pop_tag_attribute_layer("body");
}

void CH2L::reset_attribute_stack() 
{ 
	_clear_attribute_stack(); 
	_new_attribute_layer(""); 
	add_attribute(IA_PAR_ALIGN,Attribute::STRING_TYPE,DEFAULT_DOCUMENT_ALIGN);
	add_attribute(IA_FONT_NAME,Attribute::STRING_TYPE,DEFAULT_DOCUMENT_FONT);
	add_attribute(IA_FONT_SIZE,Attribute::DOUBLE_TYPE,DEFAULT_FONT_SIZE,"");
	add_attribute(IA_FONT_COLOR,Attribute::STRING_TYPE,DEFAULT_FONT_COLOR);
	add_attribute(IA_FONT_SHAPE,Attribute::INT_TYPE,0,"0");
	add_attribute(IA_FONT_SERIES,Attribute::INT_TYPE,0,"0");

	add_attribute("border",Attribute::INT_TYPE,0,"0");
	mStyleSheetRoot = StyleEntry();
}

void CH2L::html_close()
{
	dump_text();

	while (mLastDtag.size())
		end_dl();

	_implied_paragraph_close();
	make_basefont_end();

	_pop_tag_attribute_layer("body");
	_pop_tag_attribute_layer("");
	if (mAttributeStack.size())
		{
		cerr << "Warning: Attribute stack not empty" << endl;
		cerr << "**************************" << endl;
		dump_attribute_stack();
		cerr << "**************************" << endl;
		}

	reset_attribute_stack();
	mVerbatimMode = false;
	mOrderedListDepth = 0;
}

void CH2L::generate_latex_conclusion()
{
	if (!mSpidering)
		{
		//            |
		//............v.. color bracket
		cout << "\\par}" << endl << "\\end{document}" << endl;
		dump_spider_table();
		}
}

void CH2L::cleanup()
{
	dump_text();
}

void CH2L::_filter_text(const char *c,string &accum)
{
	assert(c);
	while (*c)
		{
		int i = (unsigned char)(*c);
		if (i < 0) i = 0;
		if (i > 255) i = 255;
		if (mVerbatimMode)
			{
			switch(i)
				{
				case '\t':
					accum += "\\ \\ \\ \\ \\ \\ \\ \\ ";
					break;
				case '\n':
					accum += "\\par\n";
					break;
				case '\r': // if in verbatim mode, ignore DOS newlines
					break;
				default:
					accum += gsCharTable[i];
					break;
				}
			}
		else
			{
			string newstuff = (((mWaitingForListItem /* || mWaitingForTableElement*/) && i == ' ') ? string(" ") : gsCharTable[i]);
			accum += newstuff;
			}
		c++;
		}
}

void CH2L::accumulate_text(const string &s)
{
	accumulate_text(s.c_str());
}

void CH2L::accumulate_text(const char *c)
{
	_filter_text(c,mTextAccumulator);
}

void CH2L::clear_text()
{
	mTextAccumulator="";
}

bool CH2L::dump_text()
{	
	bool result = (mTextAccumulator != "");
	cout << mTextAccumulator;
	mTextAccumulator="";

	return result;
}

void CH2L::set_title(const string &s)	
{ 
	mTitle = s; 
}


bool CH2L::causes_paragraph_reset(const string &s)
{
	return (mParagraphResetTags.find(s) != mParagraphResetTags.end());
}

void CH2L::execute_implied_closes(const string &next_tag)
{
	if (causes_paragraph_reset(next_tag))
		_implied_paragraph_close();
}

	string attribute_to_alignment_string(const Attribute &a,const string &deflt)
	{
		string result = deflt;
		string s = a.getStringVal();
		if (strcasecmp(s.c_str(),"left") == 0)
			result="flushleft";
		else if (strcasecmp(s.c_str(),"right") == 0)
			result="flushright";
		else if (strcasecmp(s.c_str(),"center") == 0)
			result="center";
		else if (strcasecmp(s.c_str(),"justify") == 0)
			result="justify";
		else
			result="justify";

		return result;
	}

			bool CH2L::_adjust_alignment(const string &s,bool force)
			{
			bool result = false;

			if ((mAlignmentString != s) || force)
				{
				cout << "\\" << s << endl;
				result = true;
				}
			mAlignmentString = s;
			return result;
			}

			void CH2L::_adjust_color(const string &color,bool force)
			{
			if ((mLastColorString != color) || force)
				cout << "\\color" << color;
			mLastColorString = color;
			}

			void CH2L::_adjust_fontsize(double v,bool force)
			{
			if ((mLastFontSize != v) || force)
				cout << "\\fontsize{" << v << "}{" << (v + (v *.2)) << "}\\selectfont ";
			mLastFontSize = v;
			}

			void CH2L::_adjust_fontname(const string &n,bool force)
			{
			if ((mLastFontName != n) || force)
				cout << "\\fontfamily{" << n << "}\\selectfont ";
			mLastFontName = n;
			}

			void CH2L::_adjust_fontshape(const string &s,bool force)
			{
			if ((mLastFontShape != s) || force)
				cout << "\\" << s << " ";
			mLastFontShape = s;
			}

			void CH2L::_adjust_fontseries(const string &s,bool force)
			{
			if ((mLastFontSeries != s) || force)
				cout << "\\" << s << " ";		
			mLastFontSeries = s;
			}

	void CH2L::_set_fontcolor()
	{
		Attribute a;
		if (_find_attribute(IA_FONT_COLOR,a))
			{
			_adjust_color(_html_color_to_latex_color(a.getStringVal()),false);
			}
	}

	void CH2L::_set_fontsize()
	{
		Attribute a;
		if (_find_attribute(IA_FONT_SIZE,a))
			{
			_adjust_fontsize(a.getDoubleVal(),false);
			}
	}


	void CH2L::_set_fontname()
	{
		Attribute a;
		if (_find_attribute(IA_FONT_NAME,a))
			_adjust_fontname(a.getStringVal(),false);
		else
			_adjust_fontname(DEFAULT_DOCUMENT_FONT,true);
	}

	void CH2L::_set_fontshape()
	{
		Attribute a;
		if (_find_attribute(IA_FONT_SHAPE,a))
			{
			if (a.getIntVal() == 0)
				_adjust_fontshape("upshape",false);
			if (a.getIntVal() & 1)
				_adjust_fontshape("itshape",false);
			if (a.getIntVal() & 2)
				_adjust_fontshape("slshape",false);
			if (a.getIntVal() & 4)
				_adjust_fontshape("scshape",false);
			}	
		else
			{		
			_adjust_fontshape("upshape",true);
			}
	}


	void CH2L::_set_fontseries()
	{
		Attribute a;
		if (_find_attribute(IA_FONT_SERIES,a))
			{
			switch(a.getIntVal())
				{
				case 0:
					_adjust_fontseries("mdseries",false);
					break;
				case 1:
					_adjust_fontseries("bfseries",false);
					break;
				default:
					_adjust_fontseries("bfseries",false);
				}
			}	
		else
			{		
			_adjust_fontseries("mdseries",true);
			}
	}

	bool CH2L::_set_fontalignment()
	{
	Attribute a;
	string alignment_string="justify";

	if (_find_attribute(IA_PAR_ALIGN,a))
		alignment_string = attribute_to_alignment_string(a,DEFAULT_DOCUMENT_ALIGN);

	return _adjust_alignment(alignment_string,false);
	}

	void CH2L::_update_alignment()
	{
	Attribute a;

	if (_find_attribute(IA_PAR_ALIGN,a))
		mAlignmentString = attribute_to_alignment_string(a,DEFAULT_DOCUMENT_ALIGN);
	}

	void CH2L::_change_alignment(const string &def,bool force)
	{
	Attribute a;
	string alignment_string=def;

	if (_find_attribute(IA_PAR_ALIGN,a))
		alignment_string = attribute_to_alignment_string(a,alignment_string);

	_adjust_alignment(alignment_string,force);
	}


	void CH2L::_set_font_all()
	{
		bool alignment_changed;
		_set_font_all(alignment_changed);
	}

	void CH2L::_set_font_all(bool &alignment_changed)
	{
		_set_fontseries();
		_set_fontshape();
		_set_fontname();
		_set_fontsize();
		_set_fontcolor();
		alignment_changed = _set_fontalignment();
	}

///
void CH2L::_font_fromstyle(const string &sss)
{
	string s = lowerstring(stripstring(sss));
	if (s == "medium")
		_siz_font_set(mRelativeFontList[4]);
	else if (s == "small")
		_siz_font_set(mRelativeFontList[3]);
	else if (s == "x-small")
		_siz_font_set(mRelativeFontList[2]);
	else if (s == "xx-small")
		_siz_font_set(mRelativeFontList[1]);
	else if (s == "large")
		_siz_font_set(mRelativeFontList[5]);
	else if (s == "x-large")
		_siz_font_set(mRelativeFontList[6]);
	else if (s == "xx-large")
		_siz_font_set(mRelativeFontList[7]);
	else if (s == "larger")
		{
		_relative_font_change(1);
		}
	else if (s == "smaller")
		{
		_relative_font_change(-1);
		}
	else if (s.size())
		{
		bool pica = false;
		bool inch = false;
		bool cm = false;
		bool mm = false;

		bool offset = (s[0] == '-' || s[0] == '+');
		if (offset)
			s = s.substr(1,s.size() - 1);
		double val = strtod(s.c_str(),0);

		if (s.find("pc") != string::npos)
			val *= 12;
		else if (s.find("in") != string::npos)
			val *= 72;
		else if (s.find("cm") != string::npos)
			val *= (72 / 2.54);
		else if (s.find("mm") != string::npos)
			val *= (72 / 25.4);
		if (offset)
			_siz_font_change(val);	
		else
			_siz_font_set(val);
		}
}

void CH2L::_siz_font_set(double val)
{
	add_attribute(IA_FONT_SIZE,Attribute::DOUBLE_TYPE,val,"");
}

void CH2L::_siz_font_change(double val)
{
	double current_size = DEFAULT_FONT_SIZE;

	Attribute a;
	if (_find_attribute(IA_FONT_SIZE,a))
		current_size = a.getDoubleVal();

	_siz_font_set(current_size + val);
}

int CH2L::_relative_font_index() const
{
	double current_size = DEFAULT_FONT_SIZE;

	Attribute a;
	if (_find_attribute(IA_FONT_SIZE,a))
		current_size = a.getDoubleVal();

	int index = 4;
	for (size_t x = 0; x < mRelativeFontList.size(); ++x)
		{
		if (mRelativeFontList[x] >= current_size)
			{
			index = x;
			break;
			}
		}

	if (index < 0)
		index = 0;
	if (index >= (int)mRelativeFontList.size())
		index = mRelativeFontList.size() - 1;

	return index;
}

void CH2L::_relative_font_change(int offset)
{
	int index = _relative_font_index() + offset;
	
	if (index < 0)
		index = 0;
	if (index >= (int)mRelativeFontList.size())
		index = mRelativeFontList.size() - 1;

	_siz_font_set(mRelativeFontList[index]);
}

void CH2L::_super_imply_close(const string &delimiter)
{
	_implied_span_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_div_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_dfn_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_samp_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_big_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_small_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_bold_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_strong_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_blink_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_italic_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_abbr_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_cite_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_em_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_address_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_acronym_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_code_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_var_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_kbd_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_tt_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_ins_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_del_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_strike_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_s_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_u_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_nobr_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_font_end(delimiter,ATTRIBUTENESTLIMIT);
	_implied_center_end(delimiter,ATTRIBUTENESTLIMIT);
}

////////////// paragraphs /////////////////


void CH2L::_implied_paragraph_close()
{
	if (dump_text()) cout << endl;

	// remove the last paragraphs tag, if any
	if (_pop_tag_attribute_layer("p"))
		{
		_set_font_all();
		_change_alignment(DEFAULT_DOCUMENT_ALIGN,false);
		}
}

void CH2L::make_paragraph_start()
{
	if (dump_text()) cout << endl;

	_search_and_apply_stylesheet();
	_alter_attribute_name("align",IA_PAR_ALIGN);
	_change_alignment(DEFAULT_PARAGRAPH_ALIGN);
	_css_setup();
	_set_font_all();
}

void CH2L::make_paragraph_end()
{
	_pop_attribute_layer();

	_implied_paragraph_close();
}

//////////////////////////////////////////////////////

void CH2L::make_input()
{
	dump_text();

	int size = 0;
	string value;
	bool checked = false;

	Attribute a;
	checked = _get_attribute("checked",a);

	if (_get_attribute("size",a))
		size = a.getIntVal();

	if (_get_attribute("value",a))
		value = a.getStringVal();

	if (_get_attribute("type",a))
		{
		string t = lowerstring(stripstring(a.getStringVal()));
		if (t == "text")
			cout << "\\nmrhtmlforminputtext" << "[" << size << "]{" << value << "}";
		else if (t == "checkbox")
			cout << "\\nmrhtmlforminputcheckbox{" << (checked ? 1 : 0) << "}";
		else if (t == "radio")
			cout << "\\nmrhtmlforminputradio{" << (checked ? 1 : 0) << "}";
		else if (t == "submit" || t == "reset" || t == "button")
			{
			if (value == "" && t == "submit")
				value="Submit";
			else if (value == "" && t == "reset")
				value="Reset";
			cout << "\\nmrhtmlforminputbutton{" << value << "}";
			}
		}

	_pop_attribute_layer();
}

void CH2L::make_image()
{
	// for now we ignore height and width because the concept
	// of pixels does not translate to a TeX environment
	// perhaps we could assume 72dpi and calculate from there?

	int image_width,image_height;
	double image_xResolution,image_yResolution;
	bool force_width = true;
	bool force_height = true;
	bool align_left = false;
	bool align_right = false;
	bool percent_height = false;
	bool percent_width = false;

	dump_text();

#ifdef ALLOW_CELL_ALIGNMENT
	if (mTableStateStack.size())
		mTableStateStack.top().mColumnNeedsABlank = false;
#endif

	string alt_text = "IMAGE ERROR";

	Attribute a;

	if (_get_attribute("alt",a))
		alt_text = a.getStringVal();

	alt_text = (mProcessImages ? (string("[") + alt_text + "]") : "");

	if (_get_attribute("align",a))
		{
		if (lowerstring(a.getStringVal()) == "left")
			align_left = true;
		else if (lowerstring(a.getStringVal()) == "right")
			align_right = true;
		}

	//if (align_left || align_right)
	// don't hypertag images because they screw up tight table aligns
	_implied_a_end("",ATTRIBUTENESTLIMIT); // not literally but latex will not allow wrapped images in an href

	if (_get_attribute("src",a) && mProcessImages)
		{
		string source = stripstring(a.getStringVal());
		string image_server,image_path,image_file;
		split_url(source,image_server,image_path,image_file);

		if (image_server == "")
			image_server = mURLServerRoot;
		if (image_path == "" || image_path[0] != '/')
			image_path = mURLCurrentPath + image_path;
		
		source = image_server + image_path + image_file;
		
		char tplate[1024];
		string tsp = mTempPath + "imgXXXXXXXX";
		strncpy(tplate,tsp.c_str(),1023);
		tplate[1023] = '\0';
		int fd = mkstemp(tplate);	// -1 if error... what to do?
		if (fd == -1)
			{
			accumulate_text("[TMP FILE CREATION ERROR]");
			goto panic;
			}
		close(fd);
		string fn = tsp = tplate;

		string::size_type suffix = source.rfind(".");
		if (suffix != string::npos)
			tsp = tsp + source.substr(suffix,source.size() - suffix);

		// convert to pdf then eps because image magic seems to have a bug
		string fn1 = tsp;
		string fn2 = fn + ".pdf";
		string fn3 = fn + ".eps";

		int result = 0;
		// check the cache
		ImageCache_t::const_iterator i_image_cache = mImageCache.find(source);
		if (i_image_cache != mImageCache.end())
			{
			const ImageDescriptor &id = (*i_image_cache).second;
			fn3 				= id.mFile;
			image_width			= id.mWidth;
			image_height		= id.mHeight;
			image_xResolution 	= id.mXResolution;
			image_yResolution 	= id.mYResolution;
			}
		else
			{
			// get the image
			cerr << "Fetching image: " << source;
			if (lowerstring(stripstring(source)).substr(0,7) == "file://" && (!mAllowFileFetch))
				source = "http://nothing";

			string::size_type spam = source.find("ad.doubleclick.net");

#ifdef LINUX
			string command;
			if (source.find("file://") == 0)
				{
				command = string("cp ") + " \"" + source.substr(7,source.size() - 7) + "\" " + fn1;
				}
			else
				{
				string base_command="wget -q -T 30 -O ";
				command = base_command + fn1 + " \"" + source + "\"";
				}
#else
			string base_command="fetch -q -T 30 -o ";
			string command = base_command + fn1 + " \"" + source + "\"";
#endif
			if (spam == string::npos)
				result = system(command.c_str());

			if (result)
				cerr << " [ERROR]" << endl;
			else if (spam != string::npos)
				{
				result = -1;
				cerr << " [SPAM]" << endl;
				}
			else
				cerr << " [ok]" << endl;

			// get image dimensions using ImageMagick
			if (result == 0)
				{
				try {
					Image image(fn1+"[0]");
					image_xResolution = image.xResolution();
					image_yResolution = image.yResolution();
					image_width = image.size().width();
					image_height = image.size().height();
					if (image_xResolution == 0) image_xResolution = 72;
					if (image_yResolution == 0) image_yResolution = 72;
					image.classType(DirectClass);
					image.write(fn3);

					ImageDescriptor id(fn3,image_width,image_height,image_xResolution,image_yResolution);
					mImageCache[source] = id;
					}
				catch(Exception &err)
					{
					cerr << "Error parsing image [" << source << "] [" << fn1 << "]" << endl;
					result = -1;
					}
				}
			}

		if (_get_attribute("width",a))
			{
			force_width = true;
			image_width = a.getIntVal();
			if (a.getAttributeType() == Attribute::PERCENT_TYPE)
				percent_width = true;
			else
				force_height = true; // if not percent do not allow perspective
			}
		if (_get_attribute("height",a))
			{
			force_height = true;
			image_height = a.getIntVal();
			if (a.getAttributeType() == Attribute::PERCENT_TYPE)
				percent_height = true;
			else
				force_width = true;	// must do both or LaTex will enlarge both the retain perspective
			}

		unlink(fn1.c_str());
		unlink(fn2.c_str());
		if (result == 0)
			{
			if (align_right || align_left)
				{
				cout << "\\endtrivlist" <<endl;
				cout << "\\makeatletter" << endl;
				cout << "\\csname @parboxrestore\\endcsname" << endl;
				cout << "\\makeatother" << endl;
				cout << "\\unskip" << endl;
				cout << "\\begin{wrapfigure}{" << (align_right ? "r" : "l") <<  "}";
				if (percent_width)
					cout << "{" << ((double)image_width) / 100.0 << "\\textwidth" << "}" << endl;
				else
					cout << "{" <<  (((double)image_width) / image_xResolution) << "in}" << endl;
				}
#ifdef ALLOW_CELL_ALIGNMENT
			//cout << "\\hbox{";
#endif
			cout << "\\includegraphics";
			if (force_height || force_width)
				{
				bool comma = false;

				cout << "[clip,";
				if (force_height)
					{
					// Note: -1 in width and height output is to compensate for LaTex rounding errors.
					// If the graphic is put in a table cell the same size, sometimes LaTeX thinks the graphic
					// is a tad bigger than the cell and wraps to a new line.
					// This was changed to 0.5 as a workaround for a latex bug that caused bad PDF generation if
					// a dimension is zero
					cout << (comma ? " , " : "") << "height=";
					if (percent_height)
						cout << ((double)image_height) / 100.0 << "\\textheight";
					else
						cout << ((((double)image_height)-0.5)) << "bp"; //cout << (((double)image_height) / image_yResolution) << "in";

					comma = true;
					}
				if (force_width)
					{
					cout << (comma ? " , " : "") << "width=";
					if (percent_width)
						cout << ((double)image_width) / 100.0 << "\\textwidth";
					else
						cout << ((((double)image_width)-0.5)) << "bp"; //cout << (((double)image_width) / image_xResolution) << "in";
					comma = true;
					}
				cout << "]";
				}
			cout << "{" << fn3 << "}";
#ifdef ALLOW_CELL_ALIGNMENT
			//cout << "}";
#endif
			if (align_right || align_left)
				cout << endl << "\\end{wrapfigure}" << endl;
			}
		else
			{
			unlink(fn3.c_str());
			accumulate_text(alt_text);
			}
		}
	else
		{
		accumulate_text(alt_text);
		}

panic:
	dump_text(); // incase 'alt' material exists
	_pop_attribute_layer();
}

void CH2L::make_linebreak()
{
	dump_text();

	_implied_a_end("",ATTRIBUTENESTLIMIT); // not literally but latex will not allow linebreaks or par breaks in an href
	_implied_u_end("",ATTRIBUTENESTLIMIT); // not literally but the uline command does not allow par breaks
	Attribute a;
	if (_get_attribute("clear",a))
		cout << "\\fillwrapfig ";

	cout << "\\par" << endl;;

	_pop_attribute_layer();
}

void CH2L::make_basefont_start()
{
	dump_text();

	// use _get and not _find since we're only interested in what this
	// tag has to say

	Attribute a;
	if (_get_attribute("size",a))
		{
		if (a.getAttributeType() == Attribute::INT_TYPE)
			{
			int v = a.getIntVal() + 1;
			if (v < 0) v = 0;
			if (v >= (int)mRelativeFontList.size())
				v = mRelativeFontList.size() - 1;
			_siz_font_set(mRelativeFontList[v]);
			}
		else if (a.getAttributeType() == Attribute::OFFSET_TYPE)
			{
			_relative_font_change(a.getIntVal());
			}
		}

	_set_fontsize();
}

void CH2L::make_basefont_end()
{
	dump_text();

	// kinda wierd. A single close basefont cancels ALL basefonts.
	// so purge until done
	_pop_tag_attribute_layer("/basefont");
	while (_pop_tag_attribute_layer("basefont"));

	_set_fontsize();
}


void CH2L::make_spacer()
{
	dump_text();

	_alter_attribute_name("align",IA_SPCR_ALIGN);

	Attribute a;
	if (_get_attribute("type",a))
		{
		string spacertype = a.getStringVal();
		if (spacertype == "horizontal")
			{
			if (_get_attribute("size",a))
				{
				double siz = a.getIntVal();
				cout << "\\hbox spread " << (siz / 72.0) << "in{}";
				}
			} // horizontal
		else if (spacertype == "vertical")
			{
			if (_get_attribute("size",a))
				{
				double siz = a.getIntVal();
				cout << "\\vbox spread " << (siz / 72.0) << "in{}";
				}
			} //vertical
		else if (spacertype == "block")
			{
			if (_get_attribute("width",a))
				{
				double width = a.getIntVal();
				if (_get_attribute("height",a))
					{
					double height = a.getIntVal();
					if (_get_attribute(IA_SPCR_ALIGN,a))
						{
						string align = a.getStringVal();
						string atag = "c";
						if (align == "top") atag = "t";
						else if (align == "texttop") atag = "t";
						else if (align == "middle") atag = "c";
						else if (align == "absmiddle") atag = "c";
						else if (align == "baseline") atag = "b";
						else if (align == "bottom") atag = "b";
						else if (align == "absbottom") atag = "b";
						cout << "\\parbox[" << atag << "]{" << (width / 72.0) << "in}{\\vbox spread " << (height / 72.0) << "in{\\hbox spread " << (width / 72.0) << "in{}}}";
						}
					}
				}
			} // block
		}

	_pop_attribute_layer();
}

void CH2L::_start_heading_style(double v)
{
	dump_text();

	add_attribute(IA_FONT_SIZE,Attribute::DOUBLE_TYPE,v,"");
	_change_alignment(DEFAULT_HEADING_ALIGN);
	_set_fontsize();
}

void CH2L::_end_heading_style(const string &tag)
{
	dump_text();

	_implied_a_end(tag,ATTRIBUTENESTLIMIT);

	_pop_tag_attribute_layer(tag);
	_change_alignment(DEFAULT_DOCUMENT_ALIGN);
	_set_fontsize();
}


void CH2L::make_horizontal_rule()
{
	if (dump_text())
		cout << endl;

	_alter_attribute_name("align",IA_PAR_ALIGN);

	_change_alignment(DEFAULT_RULE_ALIGN);

	Attribute a;

	cout << "\\unskip";
	int default_size = 1;
	if (_find_attribute("size",a))
		default_size = a.getIntVal();

	if (_find_attribute("width",a))
		{
		Attribute::Attribute_t t = a.getAttributeType();
		if (t == Attribute::PERCENT_TYPE)
			{
			cout << "\\nmrrule{" << default_size << "pt}{" << (a.getIntVal() / 100.0) << "}";
			}
		else if (t == Attribute::INT_TYPE)
			{
			cout << "\\nmrsizerule{" << default_size << "bp}{" << a.getIntVal() << "pt}";
			}
		}
	else
		{	
		cout << "\\nmrrule{" << default_size << "pt}{1}";
		}

	_pop_attribute_layer();
	_change_alignment(DEFAULT_DOCUMENT_ALIGN);
	//cout << "\\unskip ";
}


	static vector<string> parse_face_string(const string &s)
	{
		vector<string>	result;

		string accum;
		string::const_iterator i = s.begin();
		while (i != s.end())
			{
			if (isalnum(*i))
				accum += tolower(*i);
			else if (isspace(*i))
				accum += *i;
			else if ((*i) == ',' && accum != "")
				{
				result.push_back(stripstring(accum));
				accum = "";
				}
			++i;
			}
		if (accum != "")
			result.push_back(stripstring(accum));

		return result;
	}

	string CH2L::_get_tag_for_font(const string &s)
	{
		FontNameFaceMap_t::const_iterator i = mFontNameFaceMap.find(s);
		if (i != mFontNameFaceMap.end())
			return (*i).second;
		else 
			return "";
	}



void CH2L::_css_setup()
{
	Attribute a;
	int shape = -1;

	if (_get_attribute("text-align",a))
		{
		add_attribute(IA_PAR_ALIGN,Attribute::STRING_TYPE,a.getStringVal());
		}

	if (_get_attribute("font-style",a))
		{
		const string &s = a.getStringVal();
		if (s == "normal")
			shape = 0;
		else if (s == "italic")
			shape = 1;
		else if (s == "oblique")
			shape = 2;		
		}
	if (_get_attribute("font-variant",a))
		{
		if (a.getStringVal() == "small-caps")
			shape = (shape == -1 ? 4 : shape + 4);
		}

	if (shape != -1)
		add_attribute(IA_FONT_SHAPE,Attribute::INT_TYPE,shape,"");

	if (_get_attribute("font-weight",a))
		{
		const string &s = a.getStringVal();
		if (s == "normal")
			add_attribute(IA_FONT_SERIES,Attribute::INT_TYPE,0,"0");
		else if (s == "bold" || s == "bolder")
			add_attribute(IA_FONT_SERIES,Attribute::INT_TYPE,1,"1");
		}

	if (_get_attribute("font-size",a))
		{
		if (a.getAttributeType() == Attribute::STRING_TYPE)
			_font_fromstyle(a.getStringVal());
		}

	if (_get_attribute("font-family",a))
		{
		vector<string> v = parse_face_string(a.getStringVal());
		bool gotit=false;
		for(vector<string>::const_iterator i = v.begin(); i != v.end(); ++i)
			{
			string s = _get_tag_for_font(*i);
			if (s != "")
				{ add_attribute(IA_FONT_NAME,Attribute::STRING_TYPE,s); _set_fontname(); gotit = true; }

			if (gotit) break;
			}
		if (!gotit)
			{ add_attribute(IA_FONT_NAME,Attribute::STRING_TYPE,DEFAULT_DOCUMENT_FONT); _set_fontname();}
		}
	if (_get_attribute("color",a))
		{
		_alter_attribute_name("color",IA_FONT_COLOR);
		}
}


void CH2L::make_font_start()
{
	dump_text();

	// use _get and not _find since we're only interested in what this
	// tag has to say

	Attribute a;
	if (_get_attribute("size",a))
		{
		if (a.getAttributeType() == Attribute::INT_TYPE)
			{
			int v = a.getIntVal() + 1;
			if (v < 0) v = 0;
			if (v >= (int)mRelativeFontList.size())
				v = mRelativeFontList.size() - 1;
			_siz_font_set(mRelativeFontList[v]);
			}
		else if (a.getAttributeType() == Attribute::OFFSET_TYPE)
			_relative_font_change(a.getIntVal());
		}

	if (_get_attribute("face",a))
		{
		vector<string> v = parse_face_string(a.getStringVal());
		bool gotit=false;
		for(vector<string>::const_iterator i = v.begin(); i != v.end(); ++i)
			{
			string s = _get_tag_for_font(*i);
			if (s != "")
				{ add_attribute(IA_FONT_NAME,Attribute::STRING_TYPE,s); _set_fontname(); gotit = true; }

			if (gotit) break;
			}
		if (!gotit)
			{ add_attribute(IA_FONT_NAME,Attribute::STRING_TYPE,DEFAULT_DOCUMENT_FONT); _set_fontname();}
		}
	if (_get_attribute("color",a))
		{
		_alter_attribute_name("color",IA_FONT_COLOR);
		}

	_css_setup();
	_set_font_all();
}


void CH2L::_implied_font_end(const string &tagdelimiter="",int iterations=1)
{
	dump_text();
	int cnt = 0;

	while(iterations-- && _pop_tag_attribute_layer("font",tagdelimiter)) cnt++;
	if (cnt) 
		_set_font_all();
}

void CH2L::make_font_end()
{
	dump_text();
	_pop_tag_attribute_layer("/font");
	_implied_font_end("",1);
}

void CH2L::start_multicol()
{
	dump_text();
	Attribute a;

	int cols = 1;
	bool use_gutter = false;
	double gutter;

	if (_get_attribute("cols",a))
		cols = a.getIntVal();

	if (_get_attribute("gutter",a))
		{
		use_gutter = true;
		gutter = a.getIntVal();
		}
	cout << endl;
	if (use_gutter)
		cout << "\\setlength{\\columnsep}{" << (gutter / 72.0) << "in}" << endl;
	cout << "\\begin{multicols}{" << cols << "}" << endl;
}

void CH2L::end_multicol()
{
	dump_text();
	cout << endl;
	// needs an extra space because of a Latex bug
	cout << "\\ " << endl << "\\end{multicols}" << endl;
	_pop_tag_attribute_layer("multicol");
}

void CH2L::start_marquee()
{
	dump_text();
	cout << "\\nmrhtmlmarquee{";
}

void CH2L::end_marquee()
{
	dump_text();

	cout << "}";
	_pop_tag_attribute_layer("marquee");
}

void CH2L::make_center_start()	
{ 
	dump_text();

	// act like a div with attribute
	add_attribute(IA_PAR_ALIGN,Attribute::STRING_TYPE,"center");
	_change_alignment("center");

	// put out an unskip since <center> tag does not initiate a line break
	cout << "\\unskip ";
}

void CH2L::_implied_center_end(const string &tagdelimiter,int iterations)
{
	dump_text();
	int cnt = 0;

	while(iterations-- && _pop_tag_attribute_layer("center",tagdelimiter)) cnt++;
	if (cnt)
		{
		_change_alignment(DEFAULT_DOCUMENT_ALIGN);
		// put out an unskip since <center> tag does not initiate a line brea
		cout << "\\unskip ";
		}

}

void CH2L::make_center_end()
{ 
	_pop_tag_attribute_layer("/center");
	_implied_center_end();
}

void CH2L::start_q()
{
	dump_text();
	cout << "``";
}

void CH2L::end_q()
{
	dump_text();
	cout << "''";
	_pop_tag_attribute_layer("q");
}

void CH2L::start_sub()
{
	dump_text();

	cout << "$_{";
}

void CH2L::end_sub()
{
	dump_text();

	cout << "}$";
	_pop_tag_attribute_layer("sub");
}
void CH2L::start_sup()
{
	dump_text();

	cout << "$^{";
}

void CH2L::end_sup()
{
	dump_text();

	cout << "}$";
	_pop_tag_attribute_layer("sup");
}

////////////////////////
void CH2L::_implied_a_end(const string &tagdelimiter="",int iterations=1)
{
	dump_text();

	AttributeList_t	alist;
	Attribute a;

	while(iterations-- && _fetch_tag_attribute_layer(alist,"a",tagdelimiter)) 
		{
		_pop_tag_attribute_layer("a",tagdelimiter);
		if (_get_attribute(alist,"href",a))
			{
			if (mAllowHyperlinks)
				cout << "}";
			}
		}
}


void CH2L::make_a_end()
{
	dump_text();
	_pop_tag_attribute_layer("/a");
	_implied_a_end("",1);
}


void CH2L::make_a_start()
{
	dump_text();

	Attribute a;
	if (_get_attribute("name",a))
		{
		if (mAllowHyperlinks)
			cout << "\\hypertarget{" << a.getStringVal() << "}{}" << endl;
		}
	if (_get_attribute("href",a))
		{
		string url = stripstring(a.getStringVal());
		// if the name does not end in .html (or anything with a dot
		// dvips incorrectly makes it a local tag, so for now for a suffix
		if (url.size())
			{
			if (url[0] == '#')
				{
				//url = "\\"+url;
				}
			else
				{
				string server,path,file;
				split_url(url,server,path,file);
				// if file has #tag at end, remove it
				string::size_type poundloc = file.find("#");
				if (poundloc != 0 && poundloc != string::npos)
					file = file.substr(0,poundloc);
				/////
				if (server == ""  && can_spider_url(file))
					{
					string new_url = mURLServerRoot;
					if (path == "" || path[0] != '/')
						new_url += mURLCurrentPath;
					
					// if ../ then remove upwards otherwise we get into infinite url loop
					while (path.size() >= 3 && path.substr(0,3) == "../")
						{
						string::size_type slash = new_url.rfind("/");
						if (slash != string::npos)
							new_url = new_url.substr(0,slash);
						slash = new_url.rfind("/");
						if (slash != string::npos)
							new_url.substr(0,slash+1);
						path = path.substr(0,path.size() - 3);
						}					
					while (path.size() >= 2 && path.substr(0,3) == "./")
						{
						path = path.substr(0,path.size() - 2);
						}					
					///////
					new_url += path + file;
					string localtag = add_to_spider_table(new_url,"");
					if (localtag != "")
						url = string("#") + localtag;
					}
				}
			}
		url = cleanurl(url);
		if (mAllowHyperlinks)
				cout << "\\href{" << url  << "}{";
		}
	else
		{
		_pop_tag_attribute_layer("a");
		}
}

//////////////////////////////////////////////////////////////////
//////////////////// LISTS ///////////////////////////////////////
//////////////////////////////////////////////////////////////////
	#define MAX_ENUM_LIST_DEPTH	4

	string CH2L::_enum_list_internal_counter_string() const
	{
		static const string var[MAX_ENUM_LIST_DEPTH] = {
			"enumi",
			"enumii",
			"enumiii",
			"enumiv",
		};

		return  (mOrderedListDepth <= MAX_ENUM_LIST_DEPTH ? var[mOrderedListDepth-1] : "");
	}

	void CH2L::_enum_list_counter_reset_string(int val) const
	{
		string list_start_level = _enum_list_internal_counter_string();
		if (list_start_level == "") return;

		cout << "\\setcounter{" << list_start_level << "}{" << val << "}" << endl;
	}

	void CH2L::_enum_list_type_reset_string(const string &html_type) const
	{
		string list_start_level = _enum_list_internal_counter_string();
		if (list_start_level == "") return;

		static const string def_type[MAX_ENUM_LIST_DEPTH] = {
			"arabic",
			"alph",
			"roman",
			"Alph"
		};

		string list_type_val 	= (mOrderedListDepth <= MAX_ENUM_LIST_DEPTH ? def_type[mOrderedListDepth-1] : "");

		if (html_type != "")
			{
			if (html_type == "A")
				list_type_val = "Alph";
			else if (html_type == "a")
				list_type_val = "alph";
			else if (html_type == "I")
				list_type_val = "Roman";
			else if (html_type == "i")
				list_type_val = "roman";
			else if (html_type == "1")
				list_type_val = "arabic";
			}

		cout << "\\renewcommand{\\the" << list_start_level << "}{\\" << list_type_val << "{" << list_start_level << "}}" << endl;
	}

void CH2L::__implied_item_end(const string &tag)
{
	_pop_tag_attribute_layer(tag);
}

void CH2L::__make_item_start(const string &tag)
{
	dump_text();

	AttributeList_t alist;
	_fetch_tag_attribute_layer(alist,tag);
	__implied_item_end(tag);
	_push_attribute_layer(alist,tag);

	Attribute a;
	if (_get_attribute("type",a))
		{
		_enum_list_type_reset_string(a.getStringVal());
		}
	if (_get_attribute("value",a))
		{
		_enum_list_counter_reset_string(a.getIntVal() - 1);
		}

	cout << "\\item ";
	_pop_attribute_layer();
	mWaitingForListItem = false;
}


void CH2L::start_ol()			
{ 	
	mWaitingForListItem = true;
	_environment_start("enumerate"); 
	mOrderedListDepth++;
	Attribute a;

	cout << "\\setlength{\\itemsep}{0pt} "; 

	int 	list_start_val = 0;
	string 	list_type_val;

	if ((mOrderedListDepth <= MAX_ENUM_LIST_DEPTH) && _get_attribute("start",a))
		list_start_val = a.getIntVal() - 1;

	if ((mOrderedListDepth <= MAX_ENUM_LIST_DEPTH) && _get_attribute("type",a))
		list_type_val = a.getStringVal();

	if (mOrderedListDepth <= MAX_ENUM_LIST_DEPTH)
		{
		_enum_list_counter_reset_string(list_start_val);
		_enum_list_type_reset_string(list_type_val);
		}
}

void CH2L::end_ol()			
{ 
	_implied_li_end();
	_implied_lh_end();
	if (mWaitingForListItem)
		cout << "\\item ";
	mWaitingForListItem= false;
	_environment_end("enumerate","ol"); 
	mOrderedListDepth--;
}

void CH2L::_end_ul(const string &tag) 
{
	_implied_li_end(); 
	_implied_lh_end(); 
	if (mWaitingForListItem) 
		cout << "\\item "; 
	_environment_end("itemize",tag);
	mWaitingForListItem = false; 
}

//////////////// Definition List ///////////////////
/////////////////////////////////////////////////
void CH2L::_implied_dd_close()
{
	if (mLastDtag.size() == 0) return;

	string &dtag = mLastDtag.top();

	if (dtag == "dd")
		{
		cout << "\\end{list}" << endl;
		_pop_tag_attribute_layer("dd");
		}
}

void CH2L::_implied_dt_close()
{
	if (mLastDtag.size() == 0) return;

	string &dtag = mLastDtag.top();

	if (dtag == "dt")
		{
		_pop_tag_attribute_layer("dt");
		}
}

void CH2L::start_dl()			
{ 
	dump_text();

	mLastDtag.push("dl");

	mWaitingForListItem = true;
	// we set parsep to zero so any embeded tables do not have their rows spaced apart
	cout << "\\begin{list}{}{\\setlength{\\parsep}{0pt}\\setlength{\\leftmargin}{\\leftskip}}" << endl;
}

void CH2L::end_dl()			
{ 
	dump_text();

	_implied_dt_close();
	_implied_dd_close();

	mLastDtag.pop();

	cout << "\\end{list}" << endl;
	_pop_tag_attribute_layer("dl");
	mWaitingForListItem = false;
}


void CH2L::make_dt_start()
{
	dump_text();

	_implied_dd_close();
	_implied_dt_close();

	if (mLastDtag.size() == 0) return;

	mLastDtag.top() = "dt";

	cout << "\\item ";
	mWaitingForListItem = false;
}

void CH2L::make_dd_start()
{
	dump_text();

	_implied_dt_close();
	_implied_dd_close();

	if (mLastDtag.size() == 0) return;

	mLastDtag.top() = "dd";

	cout << "\\begin{list}{}{\\setlength{\\parsep}{0pt}}\\item ";
	mWaitingForListItem = false;
}

////////////////////////////////////////////////
////////////////////////////////////////////////

////////////////////// tables ////////////////////////

void CH2L::start_table()
{
	dump_text();

	// latex can't handle an href spanning a table
	_implied_a_end("",ATTRIBUTENESTLIMIT);

	if (!mProcessTables)
		{
		cout << "\\par" << endl;
		return;
		}

	// we must figure out if our width comes from nmrcolwidth or linewidth
	// (linewidth if we are nested directly in a list
	TagBag_t nestlist;
	nestlist.insert(nestlist.begin(),"table");
	nestlist.insert(nestlist.begin(),"blockquote");
	nestlist.insert(nestlist.begin(),"ul");
	nestlist.insert(nestlist.begin(),"ol");
	nestlist.insert(nestlist.begin(),"li");
	nestlist.insert(nestlist.begin(),"lh");
	nestlist.insert(nestlist.begin(),"dl");
	nestlist.insert(nestlist.begin(),"dt");
	nestlist.insert(nestlist.begin(),"dd");
	string prev_enclosure = _find_first_previous_tag(nestlist,true);
	string width_command;
	if (prev_enclosure == "table" || prev_enclosure == "")
		{
		width_command = "nmrmagicwidth";
		}
	else
		{
		width_command = "linewidth";
		}

	_alter_attribute_name("align",IA_TAB_ALIGN);
	// tables have their own little font description world
	add_attribute(IA_FONT_NAME,Attribute::STRING_TYPE,DEFAULT_DOCUMENT_FONT);
	add_attribute(IA_FONT_SIZE,Attribute::DOUBLE_TYPE,DEFAULT_FONT_SIZE,"");
	add_attribute(IA_FONT_SERIES,Attribute::INT_TYPE,0,"0");
	add_attribute(IA_FONT_SHAPE,Attribute::INT_TYPE,0,"0");
	add_attribute(IA_TABLE_DIM_INSTANCE,Attribute::INT_TYPE,(int)mTableDimInstanceCtr,"");

	Attribute a;

	_set_font_all();
	cout << "\\par ";

	TableState::mLastTableTag.push("table");
	mTableStateStack.push(TableState());

	double table_grow = 0;
	double table_percent_shrink=100;
	bool no_width = true;

	double bordersize = 0.5;
	double cellpadding = 1.5;
	double cellspacing = 1.5;

	if (_find_attribute("cellspacing",a))
		{
		if (a.getAttributeType() == Attribute::INT_TYPE)
			{
			cellspacing = a.getIntVal();
			}
		}
	if (_find_attribute("cellpadding",a))
		{
		if (a.getAttributeType() == Attribute::INT_TYPE)
			{
			cellpadding = a.getIntVal();
			}
		}
	if (_find_attribute("border",a))
		{
		if (a.getAttributeType() == Attribute::STRING_TYPE)
			{
			string s = lowerstring(stripstring(a.getStringVal()));
			if (s == "off" || s == "no") 
				bordersize=0;
			}
		else
			{
			bordersize *= a.getIntVal();
			}
		}
	else
		{
		bordersize = 0;
		}

	add_attribute(IA_MIN_CELLWIDTH,Attribute::INT_TYPE,(int)(2*(cellpadding + cellspacing + bordersize)));

	cout << "\\setlength{\\fboxrule}{" << bordersize << "bp}";
	cout << "\\setlength{\\fboxsep}{" << cellpadding + cellspacing << "bp}";
	if (_get_attribute("width",a))
		{
		if (a.getAttributeType()==Attribute::PERCENT_TYPE)
			{
			table_percent_shrink = a.getIntVal();
			no_width = false;
			}
		else if (a.getAttributeType() == Attribute::INT_TYPE)
			{
			table_percent_shrink = 0;
			table_grow = ((double)a.getIntVal()) / 72.0;
			no_width = false;
			}
		}

	if (_get_attribute(IA_TAB_ALIGN,a))
		{
		string align=stripstring(lowerstring(a.getStringVal()));
		if (align == "right" || align == "left")
			{
			if (no_width)
				{
				// alignments must have a width
				table_percent_shrink=66;
				}
			cout << "\\endtrivlist" << endl
				<< "~\\makeatletter" << endl
				<< "\\csname @parboxrestore\\endcsname" << endl
				<< "\\makeatother" << endl
				<< "\\unskip" << endl
				<< "~ \\begin{wrapfigure}{" << (align == "right" ? "r" : "l") << "}{\\" << width_command << " * \\real{"  << (table_percent_shrink/100.0) << "} + " << table_grow << "in}" << endl;
			}
		}

	cout << "{" << endl
		<< "\\setlength{\\nmrcolwidth}{\\" << width_command << " * \\real{" << (table_percent_shrink/100.0) << "} + " << table_grow << "in}" << endl;

	if (!mInlineTables)
		cout << "\\" << _dim_ctr_to_command(mTableDimInstanceCtr) << endl;
	else
		cout << "\\renewcommand{\\nmrtable}{" << endl;

	cout << "\\alloclen{tmplen}" << endl;
	cout << "\\alloclen{tmplend}" << endl;

	mTableDimInstanceCtr++;
	mWaitingForTableElement = true;
}

void CH2L::end_table()
{
	dump_text();

	if (!mProcessTables) 
		{
		_super_imply_close("table");
		_pop_tag_attribute_layer("table");
		cout << "\\par" << endl;
		return;
		}

	_implied_td_close();
	_implied_th_close();
	_implied_tr_close();

	if (mTableStateStack.size() == 0) return;

	const int &column_ctr = mTableStateStack.top().mMaxCols;
	cout << "}" << endl; 
	if (mInlineTables)
		cout << "{";

	// now we calculate the table widths
	TableState::TableRowWidths_t wlist = mTableStateStack.top().mRowWidths;

	// first some the total indicated percents
	// and count the number of TBD columns
	// and sum the pixel width columns
	double percent_tot = 0;
	int tbd_cols = 0;
	double pixel_tot = 0;
	for (TableState::TableRowWidths_t::const_iterator i = wlist.begin(); i != wlist.end(); ++i)
		{
		if ((*i).mWidthType == TableState::TableWidthEntry::PERCENT)
			percent_tot += (*i).mWidthValue;
		if ((*i).mWidthType == TableState::TableWidthEntry::TBD)
			tbd_cols++;
		if ((*i).mWidthType == TableState::TableWidthEntry::PIXELS)
			pixel_tot += (*i).mWidthValue;
		}

	// take the leftover percentage, and assign to tbd columns
	for (TableState::TableRowWidths_t::iterator i = wlist.begin(); i != wlist.end(); ++i)
		{
		if ((*i).mWidthType == TableState::TableWidthEntry::TBD)
			{
			(*i).mWidthType = TableState::TableWidthEntry::PERCENT;
			(*i).mWidthValue = (100.0-percent_tot) / tbd_cols;
			}
		}

	// now generate the percentages, remembering to subtract the pixel widths as appropriate

	assert(column_ctr == (int)wlist.size());

	int x = 0;

	Attribute a;

	// gcc having trouble doing this ia reference
	ostream *tout = 0;
	if (!mInlineTables)
			tout = &mTableDimensionsFile;
	else
			tout = &cout;

	if (_find_tag_attribute("table",IA_TABLE_DIM_INSTANCE,a))
		{
		if (!mInlineTables)
			mTableDimensionsFile << "\\newcommand{\\" << _dim_ctr_to_command(a.getIntVal()) << "}{";

		for (TableState::TableRowWidths_t::iterator i = wlist.begin(); i != wlist.end(); ++i,++x)
			{
			// the col
			(*tout) << "\\nmrmakecolsize{" << (x+1) << "}{";

			// size calc
			double wval = (*i).mWidthValue;
			if ((*i).mWidthType == TableState::TableWidthEntry::PERCENT)
				(*tout) << (wval/100.0) <<"\\nmrcolwidth - " << (pixel_tot/72.0) << "in ";
			else 
				(*tout) << (wval/72.0) << "in ";

			// subtract bounds
			(*tout) << "- 2\\fboxsep - " << (/* x == column_ctr - 1 */ true  ? "2" : "") << "\\fboxrule}" << endl;
			}
		if (!mInlineTables)
			mTableDimensionsFile << "}" << endl;
		}
	if (mInlineTables)
		cout << "\\nmrtable}}" << endl;


	if (_get_attribute(IA_TAB_ALIGN,a))
		{
		string align=stripstring(lowerstring( a.getStringVal()));
		if (align == "right" || align=="left")
			cout << "\\end{wrapfigure}" << endl;		
		}

	TableState::mLastTableTag.pop();
	mTableStateStack.pop();

	_pop_tag_attribute_layer("table");
	mWaitingForTableElement = false;
	_set_font_all();
	double bordersize = 0.5;
	double cellpadding = 1.5;
	double cellspacing = 1.5;
	if (_find_attribute("cellspacing",a))
		{
		if (a.getAttributeType() == Attribute::INT_TYPE)
			{
			cellspacing = a.getIntVal();
			}
		}
	if (_find_attribute("cellpadding",a))
		{
		if (a.getAttributeType() == Attribute::INT_TYPE)
			{
			cellpadding = a.getIntVal();
			}
		}
	if (_find_attribute("border",a))
		{
		if (a.getAttributeType() == Attribute::STRING_TYPE)
			{
			string s = lowerstring(stripstring(a.getStringVal()));
			if (s == "off" || s == "no")
				bordersize = 0;
			}
		else
			{
			bordersize *= a.getIntVal();
			}
		}
	else
		{
		bordersize = 0;
		}

	cout << "\\setlength{\\fboxrule}{" << bordersize << "bp}";
	cout << "\\setlength{\\fboxsep}{" << cellpadding + cellspacing << "bp}";

	TagBag_t nestlist;
	nestlist.insert(nestlist.begin(),"table");
	string prev_enclosure = _find_first_previous_tag(nestlist,true);
	if (prev_enclosure != "table")
		{
		cout << "\\smallskip" << endl;
		}
}

void CH2L::make_tr_end()
{
	if (!mProcessTables)
		{
		dump_text();
		_pop_attribute_layer();
		cout << "\\par" << endl;
		return;
		}

	_pop_attribute_layer();
	_implied_td_close();
	_implied_th_close();
	_implied_tr_close();
}

	void define_rowheights(int start,int end)
	{
	for (int x = start; x < end; ++x)
		{
		cout << "\\alloclen{rowheight" << val_to_suffix(x) << "}" << endl;
		cout << "\\setlength{\\rowheight" << val_to_suffix(x) << "}{0pt}" << endl;
		}
	}

void CH2L::make_tr_start()
{
	dump_text();

	if (!mProcessTables)
		{
		_pop_attribute_layer();
		cout << "\\par" << endl;
		return;
		}

	AttributeList_t	layer;
	if (!_fetch_attribute_layer(layer)) return;
	_pop_attribute_layer();
	_implied_td_close();
	_implied_th_close();
	_implied_tr_close();
	_push_attribute_layer(layer,"tr");

	if (mTableStateStack.size() == 0) return;

	TableState::mLastTableTag.push("tr");
	int &col_ctr = mTableStateStack.top().mColCtr;
	int &row_ctr = mTableStateStack.top().mRowCtr;
	int &max_rows = mTableStateStack.top().mMaxRows;
	TableState::TableRowHeights_t &row_heights = mTableStateStack.top().mRowHeights;

	col_ctr = 0;
	row_ctr += 1;
	if (max_rows < row_ctr)
		{
		while (((int)row_heights.size()) < row_ctr)
			row_heights.push_back(TableState::TableHeightEntry());
		define_rowheights(max_rows,row_ctr);
		max_rows = row_ctr;
		}

	mWaitingForTableElement = true;
}

void CH2L::___make_td_start(const string &tag)
{
	dump_text();

	if (!mProcessTables)
		{
		_pop_attribute_layer();
		cout << "\\par" << endl;
		return;
		}

	AttributeList_t layer;
	Attribute 		a;

	if (!_fetch_attribute_layer(layer)) return;
	_pop_attribute_layer();

	if (TableState::mLastTableTag.size() == 0) return;
	if (TableState::mLastTableTag.top() == "table") 
		{ 
		cerr << "Warning column after table tag" << endl;
		new_attributes("tr");
		make_tr_start();
		}

	_implied_td_close();
	_implied_th_close();
	_push_attribute_layer(layer,tag);

	TableState::mLastTableTag.push(tag);

#ifdef ALLOW_CELL_ALIGNMENT
	if (!_get_attribute("align",a))
		{
		add_attribute("align",Attribute::STRING_TYPE,"left");
		}
#endif
	_alter_attribute_name("align",IA_PAR_ALIGN);

	int min_cell_width = 0;
	if (_find_attribute(IA_MIN_CELLWIDTH,a))
		{
		min_cell_width = a.getIntVal();
		}

	int &col_ctr 	= mTableStateStack.top().mColCtr;
	int &row_ctr 	= mTableStateStack.top().mRowCtr;
	int &max_cols	= mTableStateStack.top().mMaxCols;
	int &max_rows 	= mTableStateStack.top().mMaxRows;
	int &cell_ctr	= mTableStateStack.top().mCellCtr;
	TableState::TableRowWidths_t 	&widths_list = mTableStateStack.top().mRowWidths;
	TableState::TableRowHeights_t	&row_heights = mTableStateStack.top().mRowHeights;
	TableState::TableRowSpanHeights_t	&row_span_heights = mTableStateStack.top().mRowSpanHeights;
	TableState::TableRowspanList_t	&rowspan_list = mTableStateStack.top().mRowSpans;
	TableState::TableRowspanList_t	&cellrowspan_list = mTableStateStack.top().mCellRowspans;
	abstractTable	&abstract_table = mTableStateStack.top().mAbstractTable;
	bool &needs_blank = mTableStateStack.top().mColumnNeedsABlank;

	needs_blank = true;
	int colspan = 1;
	int rowspan = 1;

	if (_get_attribute("rowspan",a))
		rowspan = a.getIntVal();

	while (((int)cellrowspan_list.size()) <= cell_ctr) cellrowspan_list.push_back(0);
	cellrowspan_list[cell_ctr] = rowspan;

	if (_get_attribute("colspan",a))
		colspan = a.getIntVal();

	bool use_height = false;
	int height = 0;
	if (_get_attribute("height",a))
		{
		use_height = true;
		height = a.getIntVal();
		}

	// output rowheight variables if needed
	if (max_rows < rowspan + row_ctr - 1)
		{
		while (((int)row_heights.size()) < (rowspan + row_ctr - 1))
			row_heights.push_back(TableState::TableHeightEntry());
		define_rowheights(max_rows,rowspan + row_ctr - 1);
		max_rows = rowspan + row_ctr - 1;
		}

	if (use_height && (rowspan == 1))
		{
		row_heights[row_ctr-1] = TableState::TableHeightEntry(TableState::TableHeightEntry::PIXELS,height);
		}
	else if (use_height)
		{
		row_span_heights.push_back(TableState::TableHeightSpanEntry(row_ctr-1,row_ctr-1+rowspan-1,cell_ctr,height));
		}

	// adjust col_ctr to account for rowspans
	// col_ctr is currently one less than this col (ie zero based)
	while (col_ctr >= (int)rowspan_list.size()) rowspan_list.push_back(0);
	while (rowspan_list[col_ctr] >= row_ctr) 
		{
		col_ctr++;
		if (col_ctr >= (int)rowspan_list.size()) rowspan_list.push_back(0);
		}
	///////
	int new_start_col = col_ctr; // new_start_col is zeror based


	// enter this columns rowspans
	for (int x = 0; x < colspan; ++x)
		{
		if (col_ctr + x >= (int)rowspan_list.size()) rowspan_list.push_back(0);
		rowspan_list[x + col_ctr] += rowspan;
		}
	///////

	col_ctr += colspan;
	if (col_ctr > max_cols) 
		max_cols = col_ctr;

	while (((int)widths_list.size()) < col_ctr)
		widths_list.push_back(TableState::TableWidthEntry());

	if (_get_attribute("width",a))
		{
		for (int x = 0; x < colspan; ++x)
			{
			TableState::TableWidthEntry &width_entry = widths_list[col_ctr-1-x];
			if (a.getAttributeType() == Attribute::PERCENT_TYPE)
				{
				width_entry.mWidthType = TableState::TableWidthEntry::PERCENT;
				width_entry.mWidthValue = ((double)a.getIntVal()) / colspan;
				}
			if (a.getAttributeType() == Attribute::INT_TYPE)
				{
				width_entry.mWidthType = TableState::TableWidthEntry::PIXELS;
				double w = ((double)a.getIntVal()) / colspan;
				if (w < min_cell_width)
					w = min_cell_width;
				width_entry.mWidthValue = w;
				}
			}
		}

	bool rgb_color = false;
	bool named_color = false;
	string color_spec;

	if (_find_attribute("bgcolor",a))
		{
		string c = a.getStringVal();
		c = lowerstring(stripstring(a.getStringVal()));
		if ((c != "white" && c != "ffffff" && c != "#ffffff") || mColorWhiteTables)
			_html_color_to_latex_info(a.getStringVal(),rgb_color,named_color,color_spec);
		}

	string suffix = val_to_suffix(cell_ctr);

	if (!rgb_color && !named_color)
		{
		cout << "\\alloccom{ccbox" << suffix << "}{0}" << endl;
		}
	else if (rgb_color)
		{
		cout << "\\definecolor{cbox" << suffix << "}{rgb}{" << color_spec << "}" << endl;
		cout << "\\alloccom{ccbox" << suffix << "}{cbox" << suffix << "}" << endl;
		}
	else if (named_color)
		{
		cout << "\\alloccom{ccbox" << suffix << "}{" << color_spec << "}" << endl;
		}

	string vertical_align = "c";

	if (_get_attribute("valign",a))
		{
		string s = lowerstring(stripstring(a.getStringVal()));
		if (s == "top")
			vertical_align = "t";
		else if (s == "bottom")
			vertical_align = "b";
		}

	cout << "\\alloccom{calign" << suffix << "}{" << vertical_align << "}" << endl;

	//cout << "\\nmrcol{c}{" << colspan << "}{black}{" << color_spec << "}{"; _set_fontalignment();

	if (_get_attribute("height",a) && (rowspan == 1))
		{
		int height = a.getIntVal();

		cout << "\t\\setlength{\\tmplen}{" << height << "bp}" << endl;
		cout << "\t\\setlength{\\tmplen}{\\tmplen / " << rowspan << "}" << endl;
		for (int x = row_ctr-1; x < (rowspan + row_ctr -1); ++x)
			{
			cout << "\t\\ifthenelse{\\tmplen > \\rowheight" << val_to_suffix(x) << "}{\\setlength{\\rowheight" << val_to_suffix(x) << "}{\\tmplen}}{}" << endl;
			}
		}

	abstract_table.addCell(row_ctr-1,new_start_col,rowspan,colspan,cell_ctr);
	cout << "\\allocbox{box" << val_to_suffix(cell_ctr) << "}\\sbox{\\box" << val_to_suffix(cell_ctr) << "}{\\begin{minipage}[c]{";

	safe_ostrstream dimensions;

	for (int x = new_start_col; x < (new_start_col + colspan); ++x)
		{
		if (x != new_start_col)
			dimensions << " + ";
		dimensions << "\\nmrctrtoken{tccol}{" << x+1 << "}" ;
		}
	if (colspan > 1)
		{
		dimensions << " + " << (2 * (colspan - 1)) << "\\fboxsep + " << (2 * (colspan - 1)) << "\\fboxrule";
		}

	const char *dim = dimensions.str();

	cout << dim << "}";

	cout << "\\setlength{\\nmrcolwidth}{" << dim << "}";

#ifdef ALLOW_CELL_ALIGNMENT
	_change_alignment(DEFAULT_DOCUMENT_ALIGN,true);
#else
	_set_fontalignment();
#endif

	mWaitingForTableElement = false;
}

void CH2L::_implied_tr_close()
{
	dump_text();

	if (!mProcessTables)
		{
		cout << "\\par" << endl;
		return;
		}

	if (TableState::mLastTableTag.size() == 0) return;
	if (mTableStateStack.size() == 0) return;

	string &tag = TableState::mLastTableTag.top();
	if (tag == "tr")
		{
		int &col_ctr = mTableStateStack.top().mColCtr;
		int &row_ctr = mTableStateStack.top().mRowCtr;
		int &max_cols = mTableStateStack.top().mMaxCols;
		int &max_rows = mTableStateStack.top().mMaxRows;
		int &cell_ctr = mTableStateStack.top().mCellCtr;
		TableState::TableRowspanList_t &rowspan_list = mTableStateStack.top().mRowSpans;
		abstractTable &abstract_table = mTableStateStack.top().mAbstractTable;
		TableState::TableRowspanList_t &cell_rowspans = mTableStateStack.top().mCellRowspans;
		TableState::TableRowHeights_t &row_heights = mTableStateStack.top().mRowHeights;
		TableState::TableRowSpanHeights_t &row_span_heights = mTableStateStack.top().mRowSpanHeights;

		_super_imply_close("tr");

		bool reduce_and_spew = true;
		int last_rowspan = -1;
		while (((int)rowspan_list.size() < max_cols))
			rowspan_list.push_back(1);

		for (int x = 0; x < max_cols; ++x)
			{
			if (x == 0)
				last_rowspan = rowspan_list[x];
			else if (last_rowspan != rowspan_list[x])
				{
				reduce_and_spew = false;
				break;
				}
			}
		col_ctr = 0;
		if (reduce_and_spew)
			{
			// spew it here
			cout << "%\n%reduce and spew\n%\n";

			///////// NEIL YOU ARE HERE!!!!
			for (TableState::TableRowSpanHeights_t::iterator i = row_span_heights.begin(); i != row_span_heights.end(); ++i)
				{
				const TableState::TableHeightSpanEntry &se = (*i);
				if (se.mStartRow >= 0 && se.mEndRow < ((int)row_heights.size()))
					{
					// get height of spanning box
					cout << "\t\\settoheight{\\tmplen}{\\usebox\\box" << val_to_suffix(se.mBoxNumber) << "}" << endl;
					cout << "\t\\settodepth{\\tmplend}{\\usebox\\box" << val_to_suffix(se.mBoxNumber) << "}" << endl;
					cout << "\t\\addtolength{\\tmplen}{\\tmplend}" << endl;
					//cout << "\t\\setlength{\\tmplen}{\\tmplen / " << rowspan << "}" << endl;
					cout << "\t\\setlength{\\tmplend}{" << se.mHeightValue << "bp}" << endl;
					cout << "\t\\ifthenelse{\\tmplen < \\tmplend}{\\setlength{\\tmplen}{\\tmplend}}{}" << endl;

					// subtract all single row absolutes
					vector<int> free_rows;
					for (int x = se.mStartRow; x <= se.mEndRow; ++x)
						{
						if (row_heights[x].mHeightType == TableState::TableHeightEntry::PIXELS)
							cout << "\\setlength{\\tmplen}{\\tmplen - \\rowheight" << val_to_suffix(x) << "}" << endl;
						else
							free_rows.push_back(x);
						}

					// divide out amongst remainder
					if (free_rows.size())
						{
						cout << "\\setlength{\\tmplen}{\\tmplen / " << free_rows.size() << "}" << endl;
						for (size_t x = 0; x < free_rows.size(); ++x)
							{
							cout << "\t\\ifthenelse{\\tmplen > \\rowheight" << val_to_suffix(free_rows[x]) << "}{\\setlength{\\rowheight" << val_to_suffix(free_rows[x]) << "}{\\tmplen}}{}" << endl;
							}
						}

					}
				}
			////////////////////////////////////////

			for (size_t x = 0; x < row_heights.size(); ++x)
				{
				if (row_heights[x].mHeightType == TableState::TableHeightEntry::PIXELS)
					{
					cout << "\\setlength{\\tmplen}{" << row_heights[x].mHeightValue << "bp}" << endl;
				//	cout << "\\setlength{\\rowheight" << val_to_suffix(x) << "}{" << row_heights[x].mHeightValue << "bp}" << endl;
					cout << "\t\\ifthenelse{\\tmplen > \\rowheight" << val_to_suffix(x) << "}{\\setlength{\\rowheight" << val_to_suffix(x) << "}{\\tmplen}}{}" << endl;
					}
				}			
			abstract_table.box();			

			max_rows = 0;
			cell_ctr = 0;
			row_ctr = 0;
			rowspan_list.erase(rowspan_list.begin(),rowspan_list.end());
			cell_rowspans.erase(cell_rowspans.begin(),cell_rowspans.end());
			}
		else
			{
			// keep accumulating
			}

		TableState::mLastTableTag.pop();
		_pop_tag_attribute_layer("tr");
		mWaitingForTableElement = false;
		}
}

void CH2L::_implied_td_close()
{
	dump_text();

	if (!mProcessTables)
		{
		cout << "\\par" << endl;
		return;
		}

	if (TableState::mLastTableTag.size() == 0) return;

	string &tag = TableState::mLastTableTag.top();
	if (tag == "td")
		{
		int &cell_ctr = mTableStateStack.top().mCellCtr;
		int &col_ctr = mTableStateStack.top().mColCtr;
		int &row_ctr = mTableStateStack.top().mRowCtr;
		bool &needs_blank = mTableStateStack.top().mColumnNeedsABlank;

		_super_imply_close("td");
		//cout << "\\ }" << endl; // so any closing justification doesn't error out
		if (needs_blank) 
			cout << "\\ ";
		cout << "\\end{minipage}}" << endl;

		assert(((int)mTableStateStack.top().mCellRowspans.size()) > cell_ctr);
		int rowspan = mTableStateStack.top().mCellRowspans[cell_ctr];

		cout << "\t\\settoheight{\\tmplen}{\\usebox\\box" << val_to_suffix(cell_ctr) << "}" << endl;
		cout << "\t\\settodepth{\\tmplend}{\\usebox\\box" << val_to_suffix(cell_ctr) << "}" << endl;
		cout << "\t\\addtolength{\\tmplen}{\\tmplend}" << endl;
		cout << "\t\\setlength{\\tmplen}{\\tmplen / " << rowspan << "}" << endl;
		if (rowspan == 1)
			{
			for (int x = row_ctr-1; x < (rowspan + row_ctr -1); ++x)
				{
				cout << "\t\\ifthenelse{\\tmplen > \\rowheight" << val_to_suffix(x) << "}{\\setlength{\\rowheight" << val_to_suffix(x) << "}{\\tmplen}}{}" << endl;
				}
			}

		cell_ctr++;

		TableState::mLastTableTag.pop();
		_pop_tag_attribute_layer("td");
		mWaitingForTableElement = false;
		_update_alignment();
		}
}



void CH2L::_implied_th_close()
{
	dump_text();

	if (!mProcessTables)
		{
		cout << "\\par" << endl;
		return;
		}

	if (TableState::mLastTableTag.size() == 0) return;

	string &tag = TableState::mLastTableTag.top();
	if (tag == "th")
		{
		int &cell_ctr = mTableStateStack.top().mCellCtr;
		int &col_ctr = mTableStateStack.top().mColCtr;
		int &row_ctr = mTableStateStack.top().mRowCtr;
		bool &needs_blank = mTableStateStack.top().mColumnNeedsABlank;

		_super_imply_close("th");
		if (needs_blank) 
			cout << "\\ ";
		cout << "\\end{minipage}}" << endl;

		cout << "\t\\settoheight{\\tmplen}{\\usebox\\box" << val_to_suffix(cell_ctr) << "}" << endl;
		cout << "\t\\settodepth{\\tmplend}{\\usebox\\box" << val_to_suffix(cell_ctr) << "}" << endl;
		cout << "\t\\addtolength{\\tmplen}{\\tmplend}" << endl;
		int rowspan = mTableStateStack.top().mRowSpans[col_ctr-1]; // should work for colspans as well since all entries will be the same
		cout << "\t\\setlength{\\tmplen}{\\tmplen / " << rowspan << "}" << endl;
		for (int x = row_ctr-1; x < (rowspan+row_ctr-1) ; ++x)
			{
			cout << "\t\\ifthenelse{\\tmplen > \\rowheight" << val_to_suffix(x) << "}{\\setlength{\\rowheight" << val_to_suffix(x) << "}{\\tmplen}}{}" << endl;
			}

		cell_ctr++;

		TableState::mLastTableTag.pop();
		_pop_tag_attribute_layer("th");
		mWaitingForTableElement = false;
		_update_alignment();
		}
}

///////////////////////////////// FORMS ////////////////////////

void CH2L::start_button()
{
	dump_text();

	cout << "\\nmrhtmlforminputbutton{";
}

void CH2L::end_button()
{
	dump_text();

	cout << "}";
	_pop_tag_attribute_layer("button");
}

void CH2L::start_textarea()
{
	dump_text();

	string height="\\height";
	string width="\\width";

	Attribute a;
	if (_get_attribute("cols",a))
		{
		safe_ostrstream o;
		o << a.getIntVal();
		width = o.str();
		width += "em";
		}
	if (_get_attribute("rows",a))
		{
		safe_ostrstream o;
		o << a.getIntVal();
		height = o.str();
		height += "ex";
		}
	cout << "\\nmrhtmlformtextarea{" << height << "}{" << width << "}{";
}

void CH2L::end_textarea()
{
	dump_text();

	cout << "}";
	_pop_tag_attribute_layer("textarea");
}

void CH2L::start_select()
{
	dump_text();
	cout << "\\shadowbox{SELECT}";
}

void CH2L::end_select()
{
	clear_text();
	_pop_tag_attribute_layer("select");
}

void CH2L::make_option_start()
{
	clear_text();
	_pop_attribute_layer();
}

void CH2L::make_option_end()
{
	clear_text();
	_pop_attribute_layer();
}

//////////
void CH2L::_font_series_start(int code)
{
	dump_text();
	add_attribute(IA_FONT_SERIES,Attribute::INT_TYPE,code,"");
	_set_fontseries();
}

void CH2L::_implied_font_series_end(const string &htmltag,const string &tagdelimiter,int iterations)
{
	dump_text();
	int cnt = 0;

	while(iterations-- && _pop_tag_attribute_layer(htmltag,tagdelimiter)) cnt++;
	if (cnt) 
		_set_fontseries();
}

void CH2L::_font_series_end(const string &tag)
{
	dump_text();
	_pop_tag_attribute_layer(string("/") + tag);
	_implied_font_series_end(tag,"",1);
}

void CH2L::_relative_font_size_start(int siz)
{
	dump_text();
	if (siz < 0) siz = 0;
	if (siz >= (int)mRelativeFontList.size())
		siz = mRelativeFontList.size() - 1;
	
	add_attribute(IA_FONT_SIZE,Attribute::DOUBLE_TYPE,mRelativeFontList[siz],"");
	_set_fontsize();
}

void CH2L::_implied_relative_font_size_end(const string &htmltag,const string &tagdelimiter,int iterations)
{
	dump_text();
	int cnt = 0;

	while(iterations-- && _pop_tag_attribute_layer(htmltag,tagdelimiter)) cnt++;
	if (cnt) 
		_set_fontsize();
}

void CH2L::_relative_font_size_end(const string &tag)
{
	dump_text();
	_pop_tag_attribute_layer(string("/") + tag);
	_implied_relative_font_size_end(tag,"",1);
}

void CH2L::_font_shape_start(int code)
{
	dump_text();
	add_attribute(IA_FONT_SHAPE,Attribute::INT_TYPE,code,"");
	_set_fontshape();
}

void CH2L::_implied_font_shape_end(const string &htmltag,const string &tagdelimiter,int iterations)
{
	dump_text();
	int cnt = 0;
	while(iterations-- && _pop_tag_attribute_layer(htmltag,tagdelimiter)) cnt++;
	if (cnt)
		_set_fontshape();
}

void CH2L::_font_shape_end(const string &tag)
{
	dump_text();
	_pop_tag_attribute_layer(string("/") + tag);
	_implied_font_shape_end(tag,"",1);
}

void CH2L::_font_name_start(const string &name)
{
	dump_text();
	add_attribute(IA_FONT_NAME,Attribute::STRING_TYPE,name);
	_set_fontname();
}

void CH2L::_implied_font_name_end(const string &htmltag,const string &tagdelimiter,int iterations)
{
	dump_text();
	int cnt = 0;
	while(iterations-- && _pop_tag_attribute_layer(htmltag,tagdelimiter)) cnt++;
	if (cnt)
		_set_fontname();
}

void CH2L::_font_name_end(const string &tag)
{
	dump_text();
	_pop_tag_attribute_layer(string("/") + tag);
	_implied_font_name_end(tag,"",1);
}

void CH2L::_simple_tag_start(const string &s)
{
	dump_text();

	cout << "\\" << s << "{";
}

void CH2L::_implied_simple_tag_end(const string &htmltag,const string &tagdelimiter,int iterations)
{
	dump_text();
	while(iterations-- && _pop_tag_attribute_layer(htmltag,tagdelimiter)) cout << "}";
}

void CH2L::_simple_tag_end(const string &htmltag)
{
	dump_text();
	_pop_tag_attribute_layer(string("/") + htmltag);
	_implied_simple_tag_end(htmltag,"",1);
}

bool CH2L::_idle_tag_start(bool allow_alignment)
{
	bool alignment_changed = false;

	dump_text();

	_search_and_apply_stylesheet();
	_alter_attribute_name("align",allow_alignment ? IA_PAR_ALIGN : IA_JUNKTAG);
	_css_setup();
	_set_font_all(alignment_changed);
	return alignment_changed;
}

void CH2L::_implied_idle_tag_end(const string &htmltag,const string &tagdelimiter,int iterations)
{
	bool changed_alignment;
	bool did_something;

	_implied_idle_tag_end(htmltag,tagdelimiter,iterations,did_something,changed_alignment);
}

void CH2L::_implied_idle_tag_end(const string &htmltag,const string &tagdelimiter,int iterations,bool &did_something,bool &changed_alignment)
{
	changed_alignment = false;
	did_something = false;

	dump_text();
	int cnt = 0;
	while(iterations-- && _pop_tag_attribute_layer(htmltag,tagdelimiter)) cnt++;
	if (cnt)
		{
		did_something = true;
		_set_font_all(changed_alignment);
		}
}

void CH2L::_idle_tag_end(const string &htmltag,bool &changed_alignment)
{
	bool did_something;

	dump_text();
	_pop_tag_attribute_layer(string("/") + htmltag);
	_implied_idle_tag_end(htmltag,"",1,did_something,changed_alignment);
}

void CH2L::_idle_tag_end(const string &htmltag)
{
	bool changed_alignment;

	_idle_tag_end(htmltag,changed_alignment);
}


void CH2L::_implied_div_end(const string &tagdelimiter="",int iterations=1) 
{ 
	bool did_something = false;
	bool alignment_changed = false;

	_implied_idle_tag_end("div",tagdelimiter,iterations,did_something,alignment_changed);
 	if (did_something)
		{
		if (alignment_changed)
			cout << "\\unskip "; 
		else 
			cout << "\\par "; 
		}
		
}

void CH2L::make_div_start()	
{ 
	if (_idle_tag_start(true)) 
		cout << "\\unskip "; 
	else 
		cout << "\\par ";
}

void CH2L::make_div_end()		
{ 
	bool alignment_changed = false;	
	_idle_tag_end("div",alignment_changed);
	if (alignment_changed)
		cout << "\\unskip "; 
	else 
		cout << "\\par ";
}

/////////


void CH2L::start_pre()		
{ 
	_font_name_start(mFontNameFaceMap["courier"]);
	mVerbatimMode = true; 
}

void CH2L::_end_pre(const std::string &tag) 
{ 
	_font_name_end(tag);
	_change_alignment(DEFAULT_DOCUMENT_ALIGN); 
	mVerbatimMode = false; 
}

//////////////////////////////////////////////////////

void CH2L::_environment_start(const string &e)
{
	if (dump_text())
		cout << endl;

	cout << "\\begin{" << e << "}" << endl;
}

void CH2L::_environment_end(const string &e,const string &tag)
{
	if (dump_text())
		cout << endl;

	cout << "\\end{" << e << "}" << endl;
	_pop_tag_attribute_layer(tag);
	_set_font_all();
}

/////////////////////////////////

void CH2L::_alter_attribute_name(const std::string &old_name,const std::string &new_name)
{
	AttributeList_t &al = mAttributeStack.front().second;
	AttributeList_t::iterator li = al.find(old_name);
	if (li != al.end())
		{
		Attribute a = (*li).second;
		al.erase(li);
		al[new_name] = a;
		}
}

bool CH2L::_get_attribute(const string &name,Attribute &a) const
{
	return _get_attribute(mAttributeStack.front().second,name,a);
}

bool CH2L::_get_attribute(const AttributeList_t &alist,const std::string &name,Attribute &a)const
{
	bool result = false;

	AttributeList_t::const_iterator li = alist.find(name);
	if (li != alist.end())
		{
		result = true;
		a = (*li).second;
		return result;
		}
	return result;
}

bool CH2L::_find_attribute(const string &name,Attribute &a) const
{
	bool result = false;

	for(AttributeStack_t::const_iterator i = mAttributeStack.begin(); i != mAttributeStack.end(); ++i)
		{
		const AttributeList_t &al = (*i).second;
		AttributeList_t::const_iterator li = al.find(name);
		if (li != al.end())
			{
			result = true;
			a = (*li).second;
			return result;
			}
		}

	return result;
}

bool CH2L::_find_tag_attribute(const TagBag_t &tag_bag,const string &name,Attribute &a) const
{
	bool result = false;

	for(AttributeStack_t::const_iterator i = mAttributeStack.begin(); i != mAttributeStack.end(); ++i)
		{
		const AttributeList_t &al = (*i).second;
		const string &tag = (*i).first;
		if (tag_bag.find(tag) != tag_bag.end())
			{
			AttributeList_t::const_iterator li = al.find(name);
			if (li != al.end())
				{
				result = true;
				a = (*li).second;
				return result;
				}
			}
		}

	return result;
}

bool CH2L::_find_tag_attribute(const std::string &tag,const string &name,Attribute &a) const
{
	TagBag_t bag;
	bag.insert(bag.begin(),tag);
	return _find_tag_attribute(bag,name,a);
}

void CH2L::_clear_attribute_stack()
{
	mAttributeStack.erase(mAttributeStack.begin(),mAttributeStack.end());
}

void CH2L::_pop_prior_attribute_layer()
{
	AttributeStack_t::iterator i = mAttributeStack.begin();
	++i;
	mAttributeStack.erase(i);
}

void CH2L::_pop_attribute_layer()
{
	mAttributeStack.pop_front();
}

bool CH2L::_pop_tag_attribute_layer(const string &tag,const string &delimiter)
{
	bool found = false;
	for(AttributeStack_t::iterator i = mAttributeStack.begin(); i != mAttributeStack.end(); ++i)
		{
		if (delimiter == (*i).first && (tag != "" || delimiter != ""))
			break;

		if (tag == (*i).first)
			{
			found = true;
			mAttributeStack.erase(i);
			break;
			}
		}
	return found;
}


void CH2L::_new_attribute_layer(const string &tag)
{
	mAttributeStack.push_front(pair<string,AttributeList_t>(tag,AttributeList_t()));
}

void CH2L::add_attribute(const Attribute &a)
{
	mAttributeStack.front().second[a.getName()] = a;
}

void CH2L::add_attribute(const string &name,Attribute::Attribute_t at,int intval,const std::string &string_rep)
{
	mAttributeStack.front().second[name] = Attribute(name,at,intval,string_rep);
}

void CH2L::add_attribute(const string &name,Attribute::Attribute_t at,double dval,const std::string &string_rep)
{
	mAttributeStack.front().second[name] = Attribute(name,at,dval,string_rep);
}

void CH2L::add_attribute(const string &name,Attribute::Attribute_t at,const std::string &s)
{
	mAttributeStack.front().second[name] = Attribute(name,at,s);
}

bool CH2L::_fetch_attribute_layer(AttributeList_t &l) const
{
	if (mAttributeStack.size() == 0) return false;

	l = mAttributeStack.front().second;
	return true;
}

bool CH2L::_fetch_tag_attribute_layer(AttributeList_t &l,const string &tag,const string &delimiter) const
{
	bool found = false;
	for(AttributeStack_t::const_iterator i = mAttributeStack.begin(); i != mAttributeStack.end(); ++i)
		{
		if (delimiter == (*i).first && (tag != "" || delimiter != ""))
			break;

		if (tag == (*i).first)
			{
			found = true;
			l = (*i).second;
			break;
			}
		}
	return found;
}

void CH2L::_push_attribute_layer(const AttributeList_t &layer,const std::string &tag)
{
	mAttributeStack.push_front(pair<string,AttributeList_t>(tag,layer));
}

/* search up the stack for any tag in the tag set and return the first one found */
string CH2L::_find_first_previous_tag(const TagBag_t &tags,bool start_with_previous)
{
	string result="";

	for(AttributeStack_t::const_iterator i = mAttributeStack.begin(); (i != mAttributeStack.end()) && (result==""); ++i)
		{
		if (start_with_previous && (i == mAttributeStack.begin()))
			continue;

		if (tags.find((*i).first) != tags.end())
			result = (*i).first;
		}
	return result;
}


/////////////////////////////////////////////////

bool CH2L::is_single_tag(const string &tag) const
{
	return (mSingleTags.find(tag) != mSingleTags.end());
}

////////////////// spidering ////////////////////////////

string CH2L::add_to_spider_table(const std::string &url,const std::string &tag,bool processed)
{
	if (url == "") return "";

	string mytag = tag;

	if (mytag == "")
		{
		safe_ostrstream o;
		o << "spider_xxx_" << (mSpiderTagCounter++);
		mytag = o.str();
		}

	if (mSpiderTable.find(url) == mSpiderTable.end())
		mSpiderTable[url] = spider_info(url,mytag,processed);
	else
		mytag = mSpiderTable[url].mLocalTag;

	return mytag;
}

void CH2L::next_spider_url(string &url,string &tag)
{
	url = ""; tag = "";

	for (SpiderTable_t::iterator i = mSpiderTable.begin(); i != mSpiderTable.end(); ++i)
		{
		if (!((*i).second.mProcessed))
			{
			(*i).second.mProcessed = true;
			url = (*i).second.mUrl;
			tag = (*i).second.mLocalTag;
			return;
			}
		}
}


	ostream& operator<<(ostream &o,const CH2L::spider_info &si)
	{
		o << "url=" << si.mUrl << " tag=" << si.mLocalTag << " processed=" << (si.mProcessed ? "true" : "false");
		return o;
	}

void CH2L::dump_spider_table() const
{
	cout << endl << "%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl;
	cout << "%% Spider Table" << endl;
	cout << "%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl << endl;
	for (SpiderTable_t::const_iterator i = mSpiderTable.begin();i != mSpiderTable.end(); ++i)
		{
		cout << "%% " << (*i).second << endl;
		}
}

void CH2L::dump_attribute_stack() const
{
	for (AttributeStack_t::const_iterator i = mAttributeStack.begin(); i != mAttributeStack.end(); ++i)
		{
		cerr << (*i).first << endl;
		}

}

bool CH2L::openTableDimensionsFile()
{
	if (!mInlineTables)
		{
		char tplate[1024];
		string tsp = mTempPath + "tabXXXXXXXX";
		strncpy(tplate,tsp.c_str(),1023);
		tplate[1023] = '\0';

		int fd = mkstemp(tplate);
		if (fd == -1)
			return false;
		close(fd);
		tsp = string(tplate) + ".tex";

		mTableDimensionsFileName = tsp;
		mTableDimensionsFile.open(tsp.c_str());
		return mTableDimensionsFile;
		}
	else 
		{
		return true;
		}
}

void CH2L::closeTableDimensionsFile()
{
	if (!mInlineTables)
		mTableDimensionsFile.close();
}

string	CH2L::_dim_ctr_to_command(size_t ctr)
{
	return string("nmrtblsz") + val_to_suffix(ctr);
}


void CH2L::add_style_sheet_entry(const vector<vector<StyleTag> > &tags, const AttributeList_t &alist)
{
	for (vector<vector<StyleTag> >::const_iterator i = tags.begin(); i != tags.end(); ++i)
		_add_style_sheet_locator_chain((*i),alist);
}

void CH2L::_add_style_sheet_locator_chain(const vector<StyleTag> &tags,const AttributeList_t &alist)
{
	// add them backwards
	StyleEntry *e = &mStyleSheetRoot;

	for (vector<StyleTag>::const_reverse_iterator i = tags.rbegin(); i != tags.rend(); ++i)
		{
		if (e->hasParent(*i))
			{
			e = &(e->getParent(*i));
			}
		else
			{
			e = &(e->addParent(*i));
			}
		}
	if (e != &mStyleSheetRoot)
		e->setAttributes(alist);
}

//#define TRACE_STYLESHEET 1

	pair<int,AttributeList_t>	CH2L::_score_ss_single(AttributeStack_t::const_iterator i,const StyleEntry &e)
	{
		pair<int,AttributeList_t> result;
		result.first = -1000000;

		int last_score = result.first;

		for (; i != mAttributeStack.end(); ++i)
			{
			// create a style tag from current iterator
			Attribute a;
			string tag,cls,id;
			tag = (*i).first;
			if (_get_attribute((*i).second,"class",a))
				cls = a.getStringVal();
			if (_get_attribute((*i).second,"id",a))
				id = a.getStringVal();

			StyleTag	t(tag,cls,id);

			if (e.getStyleTag().moregeneral(t) || t == e.getStyleTag())
				{
				int v = ( t == e.getStyleTag() ? 10 : 5);
				AttributeList_t alist = e.getStyleTag().getAttributes();
				if (e.getParentList().size() == 0)
					{
					if (v > result.first)
						{
						result.first = v;
						result.second = alist;
						}
					}
				else
					{
					AttributeStack_t::const_iterator ii = i;
					pair<int,AttributeList_t> p = _score_ss_path(++ii,e);
					v += p.first;
					alist = p.second;
					if (v > result.first)
						{
						result.first = v;
						result.second = alist;
						}
					}
				}


#ifdef TRACE_STYLESHEET
			cerr << "AttrStack: " << (*i).first << " Entry: ";
			e.dump(cerr,"");
			cerr << " Score: " << result.first << "/" << result.first - last_score << " Attributes: " << result.second.size() << endl;
			last_score = result.first;
#endif
			}
		return result;
	}

	pair<int,AttributeList_t>	CH2L::_score_ss_path(AttributeStack_t::const_iterator i,const StyleEntry &e)
	{
		pair<int,AttributeList_t> result;
		result.first = -1000000;

		const StyleEntry::StyleTagList_t &l = e.getParentList();
		for (StyleEntry::StyleTagList_t::const_iterator li = l.begin(); li != l.end(); ++li)
			{
			pair<int,AttributeList_t> p = _score_ss_single(i,*li);
			if (p.first > result.first)
				result = p;			
			}
		return result;
	}

// search from current attribute stack and apply
// anything found
void CH2L::_search_and_apply_stylesheet()
{
	pair<int,AttributeList_t> result = _score_ss_path(mAttributeStack.begin(),mStyleSheetRoot);

	// apply new_attributes
#ifdef TRACE_STYLESHEET
	cerr << "RESULT Score is : " << result.first << endl;
#endif
	if (result.first <= 0)
		return;

	_merge_in_attribute_list(result.second);
}

void CH2L::_merge_in_attribute_list(const AttributeList_t &alist)
{
	if (mAttributeStack.size() == 0) return;

	for (AttributeList_t::const_iterator i = alist.begin(); i != alist.end(); ++i)
		{
		mAttributeStack.front().second[(*i).first] = (*i).second;
		}
}


