/*
 * This file is part of htmltolatex (h2l) a HTML to Latex converter
 * Copyright (C) 2001 Neil Radisch (nradisch@panix.com)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file copying.txt.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */
#ifndef CH2L_H_
#include <string>
#include <map>
#include <assert.h>
#include <list>
#include <vector>
#include <set>
#include <stack>
#include <fstream>
#include "Attribute.h"
#include "StyleTree.h"
#include "abstractTable.h"

class CH2L
{
public:

private:
	std::string	mTextAccumulator;
	std::string mTitle;
	typedef std::set<std::string>	TagBag_t;
	TagBag_t						mParagraphResetTags;
	TagBag_t						mIgnoreSpaceTags;

	//////////////// attribute stack ////////////////////////
	typedef std::list<std::pair<std::string,AttributeList_t> >	AttributeStack_t;
	AttributeStack_t	mAttributeStack;
	///////////////////////////////////////////////////////

	/////////////// tag action constructs ///////////////////
	typedef void (CH2L::*makemethod_t)();
	typedef void (CH2L::*startmethod_t)();
	typedef void (CH2L::*endmethod_t)();
	typedef std::map<std::string,makemethod_t> MakeMethodMap_t;
	typedef std::map<std::string,startmethod_t> StartMethodMap_t;
	typedef std::map<std::string,endmethod_t> EndMethodMap_t;
	MakeMethodMap_t		mMakeMethodMap;
	StartMethodMap_t	mStartMethodMap;
	EndMethodMap_t		mEndMethodMap;
	TagBag_t			mSingleTags;
	/////////////////////////////////////////////////////////

	///////////////// images /////////////////////////
	struct ImageDescriptor {
		std::string	mFile;
		int			mWidth;
		int			mHeight;
		double		mXResolution;
		double		mYResolution;

		ImageDescriptor(){}
		ImageDescriptor(const std::string &f,int w,int h,double xres,double yres)
			: mFile(f),mWidth(w),mHeight(h),mXResolution(xres),mYResolution(yres){}
		};
	
	typedef std::map<std::string, ImageDescriptor> ImageCache_t;
	ImageCache_t mImageCache;

	///////////// state ////////////////////////////////
	bool	mColorWhiteTables;
	bool	mVerbatimMode;
	bool	mWaitingForListItem;
	bool	mWaitingForTableElement;

	// URL processing
	std::string mURLServerRoot;
	std::string mURLCurrentPath;
	std::string mURLFileName;

	bool	mAllowHyperlinks;
	bool	mAllowFileFetch;
	bool	mProcessImages;
	bool	mProcessTables;
	bool	mProcessStyles;
	std::string	mPaperWidth;
	std::string	mPaperHeight;
	std::string mHorizontalMargin;
	std::string mVerticalMargin;
	std::string	mTempPath;
	std::string mPathToDataFile;
	
	size_t	mOrderedListDepth;
	stack<string>	mLastDtag; // last "definition(dd dt dl)" tag seen


	struct TableState {

		struct TableWidthEntry {
			typedef enum { TBD, PERCENT, PIXELS } TableWidth_t;
			TableWidth_t	mWidthType;
			double			mWidthValue;

			TableWidthEntry(): mWidthType(TBD),mWidthValue(0){}
			TableWidthEntry(const TableWidth_t &t,const int &v) : mWidthType(t),mWidthValue(v){}
			};

		struct TableHeightEntry {
			typedef enum { TBD, PIXELS } TableHeight_t;
			TableHeight_t	mHeightType;
			double			mHeightValue;

			TableHeightEntry(): mHeightType(TBD),mHeightValue(0){}
			TableHeightEntry(const TableHeight_t &t,const int &v) : mHeightType(t),mHeightValue(v){}
			};

		struct TableHeightSpanEntry {
			int				mStartRow,mEndRow;
			int				mBoxNumber;
			double			mHeightValue;

			TableHeightSpanEntry(): mStartRow(0),mEndRow(0),mBoxNumber(0),mHeightValue(0){}
			TableHeightSpanEntry(int start,int end,int boxnumber,const int &v) : mStartRow(start),mEndRow(end),mBoxNumber(boxnumber),mHeightValue(v){}
			};

		typedef stack<string>			LastTableTag_t;
		typedef vector<TableWidthEntry>	TableRowWidths_t;
		typedef vector<TableHeightEntry>	TableRowHeights_t;
		typedef list<TableHeightSpanEntry>	TableRowSpanHeights_t;
		typedef vector<int>				TableRowspanList_t;
		int								mColCtr;
		int								mMaxCols;
		int								mRowCtr;
		int								mMaxRows;
		int								mCellCtr;
		TableRowWidths_t				mRowWidths;
		TableRowHeights_t				mRowHeights;
		TableRowSpanHeights_t			mRowSpanHeights;
		TableRowspanList_t				mRowSpans;
		TableRowspanList_t				mCellRowspans;
		static LastTableTag_t			mLastTableTag;	// last "table(table, tr, td)" tag seen
		abstractTable					mAbstractTable;
		bool							mColumnNeedsABlank;

		TableState():
			mColCtr(0),mMaxCols(0),mRowCtr(0),mMaxRows(0),mCellCtr(0),mColumnNeedsABlank(true){}
		};
	typedef stack<TableState>	TableStateStack_t;
	TableStateStack_t	mTableStateStack;

	void _implied_tr_close();
	void _implied_td_close();
	void _implied_th_close();
	///////////////////////////////////////////////////

	///////// font size constructs //////////////////
	std::vector<double>	mRelativeFontList;
	int		_relative_font_index() const;
	void	_siz_font_set(double val);
	void	_siz_font_change(double val);
	void	_relative_font_change(int offset);
	void	_font_fromstyle(const std::string &s);
	//////////////////////////////////////////////

	///////////// font face constructs /////////////
	typedef std::map<std::string,std::string>	FontNameFaceMap_t;
	FontNameFaceMap_t	mFontNameFaceMap;

	struct latex_rgb{
			double r;
			double g;
			double b;

			latex_rgb(): r(0.0),g(0.0),b(0.0){}
			latex_rgb(double pr,double pg,double pb)
				: r(pr),g(pg),b(pb){}
	};
	typedef map<string,latex_rgb> ColorMap_t;
	ColorMap_t mLegalColorNames;
	///////////////////////////////////////////////

	//////////////// spidering //////////////////
	struct spider_info {
		std::string	mUrl;
		std::string mLocalTag;
		bool		mProcessed;

		spider_info() : mProcessed(false){}
		spider_info(const std::string &url,const std::string &tag,bool processed = false)
			: mUrl(url),mLocalTag(tag),mProcessed(processed){}
		
		friend std::ostream& operator<<(std::ostream &o,const spider_info &si);
	};
	typedef std::map<string,spider_info>	SpiderTable_t;
	SpiderTable_t	mSpiderTable;
	bool			mSpidering;	// not the first page
	size_t			mSpiderTagCounter;
	std::string		mCurrentSpiderTag;
public:
	void 		next_spider_url(std::string &url,std::string &tag);
	void		dump_spider_table() const; // diagnostic
	std::string	add_to_spider_table(const std::string &url,const std::string &tag,bool processed=false);
private:
	////////////////////////////////////////////

	/////////////// enumerated list helpers /////////////////////////
	std::string _enum_list_internal_counter_string() const;
	void _enum_list_counter_reset_string(int val) const;
	void _enum_list_type_reset_string(const std::string &html_type) const;
	///////////////////////////////////////////////////////////

	void _close_minipage();
	void _start_heading_style(double val);
	void _end_heading_style(const std::string &tag);
	void _environment_start(const std::string &s);
	void _environment_end(const std::string &s,const std::string &tag);
	std::string _html_color_to_latex_color(const string &s) const;
	void _html_color_to_latex_info(const string &sin,bool &rgb,bool &named,string &spec) const;
	void _set_hyperlink_color(const string &s) const;

	std::string _get_tag_for_font(const std::string &s);

	void make_horizontal_rule();
	void make_linebreak();

	/////////// paragraphs ////////////////////
	bool causes_paragraph_reset(const std::string &s);
	void _implied_paragraph_close();
	void make_paragraph_start();
	void make_paragraph_end();
	/////////////////////////////////////////
	std::string mAlignmentString;
	bool _adjust_alignment(const std::string &s,bool force);
	void _change_alignment(const std::string &def,bool force=true);
	void _update_alignment();
	std::string mLastColorString;
	void _adjust_color(const std::string &s,bool force);
	std::string mLastFontSeries;
	void _adjust_fontseries(const std::string &s, bool force);
	std::string mLastFontShape;
	void _adjust_fontshape(const std::string &s,bool force);
	std::string mLastFontName;
	void _adjust_fontname(const std::string &s,bool force);
	double mLastFontSize;
	void _adjust_fontsize(double v,bool force);

	void _set_font_all(bool &alignment_changed); // returns extra info
	void _set_font_all();
	void _set_fontseries();
	void _set_fontshape();
	void _set_fontname();
	void _set_fontsize();
	void _set_fontcolor();
	bool _set_fontalignment();

	void make_basefont_start();
	void make_basefont_end();
	void make_image();
	void make_spacer();
	void make_tr_start();
	void make_tr_end();
	void make_td_start() { ___make_td_start("td"); }
	void ___make_td_start(const std::string &tag);
	void make_th_start() { ___make_td_start("th"); 	cout << "\\bfseries \\center ";}
	void make_input();
	void make_option_start();
	void make_option_end();

	void start_h1()		{ _start_heading_style(mRelativeFontList[7]); }
	void end_h1()		{ _end_heading_style("h1"); }
	void start_h2()		{ _start_heading_style(mRelativeFontList[6]); }
	void end_h2()		{ _end_heading_style("h2"); }
	void start_h3()		{ _start_heading_style(mRelativeFontList[5]); }
	void end_h3()		{ _end_heading_style("h3"); }
	void start_h4()		{ _start_heading_style(mRelativeFontList[4]); }
	void end_h4()		{ _end_heading_style("h4"); }
	void start_h5()		{ _start_heading_style(mRelativeFontList[3]); }
	void end_h5()		{ _end_heading_style("h5"); }
	void start_h6()		{ _start_heading_style(mRelativeFontList[2]); }
	void end_h6()		{ _end_heading_style("h6"); }
	void start_center();
	void end_center();


	void _super_imply_close(const std::string &delimiter);

	// **************************************************************************
	void make_center_start();
	void _implied_center_end(const std::string &tagdelimiter="", int iterations=1);
	void make_center_end();
	///////////////
	bool _idle_tag_start(bool allow_alignment);
	void _implied_idle_tag_end(const string &htmltag,const string &tagdelimiter,int iterations);
	void _idle_tag_end(const std::string &tag);
	void _implied_idle_tag_end(const string &htmltag,const string &tagdelimiter,int iterations,bool &did_something,bool &changed_alignment);
	void _idle_tag_end(const std::string &tag,bool &changed_alignment);
	void _implied_span_end(const string &tagdelimiter="",int iterations=1) { _implied_idle_tag_end("span",tagdelimiter,iterations); }
	void make_span_start()	{ _idle_tag_start(false); }
	void make_span_end()	{ _idle_tag_end("span"); }
	void _implied_div_end(const string &tagdelimiter="",int iterations=1);
	void make_div_start();
	void make_div_end();
	void _implied_dfn_end(const string &tagdelimiter="",int iterations=1) { _implied_idle_tag_end("dfn",tagdelimiter,iterations); }
	void make_dfn_start()	{ _idle_tag_start(false); }
	void make_dfn_end()		{ _idle_tag_end("dfn"); }
	void _implied_samp_end(const string &tagdelimiter="",int iterations=1) { _implied_idle_tag_end("samp",tagdelimiter,iterations); }
	void make_samp_start()	{ _idle_tag_start(false); }
	void make_samp_end()	{ _idle_tag_end("samp"); }

	////////////////
	void _relative_font_size_start(int siz);
	void _implied_relative_font_size_end(const string &htmltag,const string &tagdelimiter,int iterations);
	void _relative_font_size_end(const std::string &tag);

	void _implied_big_end(const string &tagdelimiter="",int iterations=1) { _implied_relative_font_size_end("big",tagdelimiter,iterations); }
	void make_big_start()	{ _relative_font_size_start(_relative_font_index() + 1); }
	void make_big_end()		{ _relative_font_size_end("big"); }
	void _implied_small_end(const string &tagdelimiter="",int iterations=1) { _implied_relative_font_size_end("small",tagdelimiter,iterations); }
	void make_small_start()	{ _relative_font_size_start(_relative_font_index() - 1); }
	void make_small_end()	{ _relative_font_size_end("small"); }
	////////////////
	void _font_series_start(int code);
	void _implied_font_series_end(const string &htmltag,const string &tagdelimiter,int iterations);
	void _font_series_end(const string &tag);

	void _implied_bold_end(const string &tagdelimiter="",int iterations=1) { _implied_font_series_end("b",tagdelimiter,iterations); }
	void make_bold_start()		{ _font_series_start(1); }
	void make_bold_end()		{ _font_series_end("b"); }
	void _implied_strong_end(const string &tagdelimiter="",int iterations=1) { _implied_font_series_end("strong",tagdelimiter,iterations); }
	void make_strong_start()	{ _font_series_start(1); }
	void make_strong_end()		{ _font_series_end("strong"); }
	void _implied_blink_end(const string &tagdelimiter="",int iterations=1) { _implied_font_series_end("blink",tagdelimiter,iterations); }
	void make_blink_start() 	{ _font_series_start(1); }
	void make_blink_end() 		{ _font_series_end("blink"); }
	////////////////////
	void _font_shape_start(int code);
	void _implied_font_shape_end(const string &htmltag,const string &tagdelimiter,int iterations);
	void _font_shape_end(const string &tag);

	void _implied_italic_end(const string &tagdelimiter="",int iterations=1) { _implied_font_shape_end("i",tagdelimiter,iterations); }
	void make_italic_start()	{ _font_shape_start(1); }
	void make_italic_end()		{ _font_shape_end("i"); }
	void _implied_abbr_end(const string &tagdelimiter="",int iterations=1) { _implied_font_shape_end("abbr",tagdelimiter,iterations); }
	void make_abbr_start()		{ _font_shape_start(1); }
	void make_abbr_end()		{ _font_shape_end("abbr"); }
	void _implied_cite_end(const string &tagdelimiter="",int iterations=1) { _implied_font_shape_end("cite",tagdelimiter,iterations); }
	void make_cite_start()		{ _font_shape_start(1); }
	void make_cite_end()		{ _font_shape_end("cite"); }
	void _implied_em_end(const string &tagdelimiter="",int iterations=1) { _implied_font_shape_end("em",tagdelimiter,iterations); }
	void make_em_start()		{ _font_shape_start(1); }
	void make_em_end()			{ _font_shape_end("em"); }
	void _implied_address_end(const string &tagdelimiter="",int iterations=1) { _implied_font_shape_end("address",tagdelimiter,iterations); }
	void make_address_start()		{ _font_shape_start(1); }
	void make_address_end()			{ _font_shape_end("address"); }
	void _implied_acronym_end(const string &tagdelimiter="",int iterations=1) { _implied_font_shape_end("acronym",tagdelimiter,iterations); }
	void make_acronym_start()		{ _font_shape_start(4); }
	void make_acronym_end()			{ _font_shape_end("acronym"); }
	///////////////////
	void _font_name_start(const std::string &name);
	void _implied_font_name_end(const string &htmltag,const string &tagdelimiter,int iterations);
	void _font_name_end(const string &tag);

	void _implied_code_end(const string &tagdelimiter="",int iterations=1) { _implied_font_name_end("code",tagdelimiter,iterations); }
	void make_code_start()		{ _font_name_start(mFontNameFaceMap["courier"]); }
	void make_code_end()		{ _font_name_end("code"); }
	void _implied_var_end(const string &tagdelimiter="",int iterations=1) { _implied_font_name_end("var",tagdelimiter,iterations); }
	void make_var_start()		{ _font_name_start(mFontNameFaceMap["courier"]); }
	void make_var_end()			{ _font_name_end("var"); }
	void _implied_kbd_end(const string &tagdelimiter="",int iterations=1) { _implied_font_name_end("kbd",tagdelimiter,iterations); }
	void make_kbd_start()		{ _font_name_start(mFontNameFaceMap["courier"]); }
	void make_kbd_end()			{ _font_name_end("kbd"); }
	void _implied_tt_end(const string &tagdelimiter="",int iterations=1) { _implied_font_name_end("tt",tagdelimiter,iterations); }
	void make_tt_start()		{ _font_name_start(mFontNameFaceMap["courier"]); }
	void make_tt_end()			{ _font_name_end("tt"); }
	//////////////////
	void _simple_tag_start(const std::string &t);
	void _implied_simple_tag_end(const string &htmltag,const string &tagdelimiter,int iterations);
	void _simple_tag_end(const std::string &htmltag);

	void _implied_ins_end(const string &tagdelimiter="",int iterations=1) { _implied_simple_tag_end("ins",tagdelimiter,iterations); }
	void make_ins_start()	{ _simple_tag_start("uline"); }
	void make_ins_end()		{ _simple_tag_end("ins"); }
	void _implied_del_end(const string &tagdelimiter="",int iterations=1) { _implied_simple_tag_end("del",tagdelimiter,iterations); }
	void make_del_start()	{ _simple_tag_start("sout"); }
	void make_del_end()		{ _simple_tag_end("del"); }
	void _implied_strike_end(const string &tagdelimiter="",int iterations=1) { _implied_simple_tag_end("strike",tagdelimiter,iterations); }
	void make_strike_start()	{ _simple_tag_start("sout"); }
	void make_strike_end()		{ _simple_tag_end("strike"); }
	void _implied_s_end(const string &tagdelimiter="",int iterations=1) { _implied_simple_tag_end("s",tagdelimiter,iterations); }
	void make_s_start()		{ _simple_tag_start("sout"); }
	void make_s_end()		{ _simple_tag_end("s"); }
	void _implied_u_end(const string &tagdelimiter="",int iterations=1) { _implied_simple_tag_end("u",tagdelimiter,iterations); }
	void make_u_start()		{ _simple_tag_start("uline"); }
	void make_u_end()		{ _simple_tag_end("u"); }
	void _implied_nobr_end(const string &tagdelimiter="",int iterations=1) { _implied_simple_tag_end("nobr",tagdelimiter,iterations); }
	void make_nobr_start()		{ _simple_tag_start("hbox"); }
	void make_nobr_end()		{ _simple_tag_end("nobr"); }
	/////////////////////////
	void _implied_font_end(const string &tagdelimiter="",int iterations=1);
	void make_font_start();
	void make_font_end();
	///////////////////////
	void _implied_a_end(const string &tagdelimiter="",int iterations=1);
	void make_a_start();
	void make_a_end();
	// ****************************************************************************

	void start_sub();
	void end_sub();
	void start_sup();
	void end_sup();
	void start_pre();
	void _end_pre(const std::string &tag);
	void end_pre()			{ _end_pre("pre"); }
	void start_listing()	{ start_pre(); }
	void end_listing()		{ _end_pre("listing"); }
	void start_xmp()		{ start_pre(); }
	void end_xmp()			{ _end_pre("xmp"); }
	void start_plaintext()	{ start_pre(); }
	void end_plaintext()	{ _end_pre("plaintext"); }
	void start_blockquote()	{ _environment_start("tightquote"); }
	void end_blockquote()	{ _environment_end("tightquote","blockquote"); }
	void start_q();
	void end_q();
	void start_marquee();
	void end_marquee();
	void start_multicol();
	void end_multicol();
	//////// lists //////////
	void __implied_item_end(const std::string &tag);
	void __make_item_start(const std::string &tag);
	void _implied_li_end() 	{ __implied_item_end("li"); }
	void make_li_start()	{ __make_item_start("li"); }
	void _implied_lh_end()	{ __implied_item_end("lh"); }
	void make_lh_start()	{ __make_item_start("lh"); }
	void start_ol();
	void end_ol();
	void start_ul()			{ _environment_start("itemize"); cout << "\\setlength{\\itemsep}{0pt} "; mWaitingForListItem=true;}
	void _end_ul(const std::string &tag);
	void end_ul()			{ _end_ul("ul"); }
	//////////////////////////
	
	////// definition lists ////
	void start_dl();
	void end_dl();
	void make_dt_start();
	void make_dd_start();
	void _implied_dd_close();
	void _implied_dt_close();
	//////////////////////////

	void start_dir()		{ start_ul(); }
	void end_dir()			{ _end_ul("dir"); }
	void start_menu()		{ start_ul(); }
	void end_menu()			{ _end_ul("menu"); }
	void start_table();
	void end_table();
	void start_button();
	void end_button();
	void start_textarea();
	void end_textarea();
	void start_select();
	void end_select();

	void start_body();
	void end_body();

	void _filter_text(const char *c,std::string &accum);

public:
	CH2L();

	void setURLServerRoot(const std::string &s) { mURLServerRoot = s; }
	void setURLCurrentPath(const std::string &s) { mURLCurrentPath = s; }
	void setURLFileName(const std::string &s) { mURLFileName = s; }

	void setCurrentSpiderTag(const std::string &s) { mCurrentSpiderTag = s; }
	void setSpidering(bool s) { mSpidering = s; }

	void setColorWhiteTables(bool c) { mColorWhiteTables = c; }
	void setAllowFileFetch(bool f) { mAllowFileFetch = f; }
	bool getAllowFileFetch() const { return mAllowFileFetch; }
	void setAllowHyperlinks(bool a) { mAllowHyperlinks = a; }
	void setProcessImages(bool process) { mProcessImages = process; }
	void setProcessTables(bool process) { mProcessTables = process; }
	void setProcessStyles(bool process) { mProcessStyles = process; }
	bool getProcessStyles() const { return mProcessStyles; }
	void setPaperWidth(const std::string &spec) { mPaperWidth = spec; }
	void setPaperHeight(const std::string &spec) { mPaperHeight = spec; }
	void setHorizontalMargin(const std::string &spec) { mHorizontalMargin = spec; }
	void setVerticalMargin(const std::string &spec) { mVerticalMargin = spec; }
	void setTempPath(const std::string &p) { mTempPath = p; }
	void setPathToDataFile(const std::string &p) { mPathToDataFile = p; }

	void generate_latex_prefix();
	void latex_preheader();
	void cleanup();
	void generate_latex_conclusion();
	void html_close();

	void accumulate_text(const char *c);
	void accumulate_text(const std::string &s);

	bool dump_text();
	void clear_text();

	void execute_implied_closes(const std::string &next_tag);

	void set_title(const std::string &s);

	//////////////// Attributes //////////////////
private:
	void _css_setup();

	bool _find_attribute(const std::string &name,Attribute &a) const;
	bool _find_tag_attribute(const std::string &tag,const std::string &name,Attribute &a) const;
	bool _find_tag_attribute(const TagBag_t &tagbag,const std::string &name,Attribute &a) const;
	bool _get_attribute(const std::string &name,Attribute &a) const;
	bool _get_attribute(const AttributeList_t &alist,const std::string &name,Attribute &a)const;
	void _merge_in_attribute_list(const AttributeList_t &alist);
	void _pop_attribute_layer();
	void _pop_prior_attribute_layer();
	bool _pop_tag_attribute_layer(const std::string &tag,const std::string &delimiter="");
	void _new_attribute_layer(const std::string &tag);
	void _clear_attribute_stack();
	void dump_attribute_stack() const;
	void _alter_attribute_name(const std::string &old_name,const std::string &new_name);
	bool _fetch_attribute_layer(AttributeList_t &l) const;
	bool _fetch_tag_attribute_layer(AttributeList_t &l,const std::string &tag,const std::string &delimiter="") const;
	void _push_attribute_layer(const AttributeList_t &layer,const std::string &tag);

	std::string _find_first_previous_tag(const TagBag_t &tags,bool start_with_previous);
public:
	void new_attributes(const std::string &tag) { _new_attribute_layer(tag); }
	void add_attribute(const std::string &name,Attribute::Attribute_t at,int intval,const std::string &string_rep="");
	void add_attribute(const std::string &name,Attribute::Attribute_t at,double dval,const std::string &string_rep="");
	void add_attribute(const std::string &name,Attribute::Attribute_t at,const std::string &s);
	void add_attribute(const Attribute &a);
	void reset_attribute_stack();
	//////////////////////////////////////////////////

	void tag_make(const std::string &t);
	void tag_start(const std::string &t);
	void tag_end(const std::string &t);

	bool tag_ignores_spaces(const std::string &tag) const;
	bool verbatim_mode() const { return mVerbatimMode; }
	bool is_single_tag(const std::string &tag) const;

private:
	bool			mInlineTables;
	std::ofstream	mTableDimensionsFile;
	std::string		mTableDimensionsFileName;
	size_t			mTableDimInstanceCtr;
	std::string		_dim_ctr_to_command(size_t ctr);
public:
	bool openTableDimensionsFile();
	void closeTableDimensionsFile();
	void setInlineTables(bool i) { mInlineTables = i;}

	std::string attribute_name_converter(const std::string &s) const;


private:
	StyleEntry	mStyleSheetRoot;
	void _add_style_sheet_locator_chain(const vector<StyleTag> &tags,const AttributeList_t &alist);
	void _search_and_apply_stylesheet();
	pair<int,AttributeList_t>	_score_ss_path(AttributeStack_t::const_iterator i,const StyleEntry &e);
	pair<int,AttributeList_t>	_score_ss_single(AttributeStack_t::const_iterator i,const StyleEntry &e);
public:
	void add_style_sheet_entry(const std::vector< std::vector <StyleTag> > &tags, const AttributeList_t &alist);
};

#endif
