/*
 * This file is part of htmltolatex (h2l) a HTML to Latex converter
 * Copyright (C) 2001 Neil Radisch (nradisch@panix.com)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file copying.txt.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */
#include "abstractTable.h"
#include "safe_ostrstream.h"
#include "util.h"
#include <stdio.h>

//#define DEBUG 1

using namespace std;


ostream& operator<<(ostream &o,const tabCell &c)
{
	o << "row=" << c.mRow << " col=" << c.mCol << " rowspan=" << c.mRowspan << " colspan=" << c.mColspan << " cell=" << c.mCellNumber;
	return o;
}

size_t	abstractTable::getRows() const
{
	size_t rows = 0;
	for(const_iterator i = begin(); i != end(); ++i)
		if (((*i).mRow+1) > rows)
			rows = (*i).mRow+1;
	return rows;
}

size_t abstractTable::getCols() const
{
	size_t cols = 0;
	for (const_iterator i = begin(); i != end(); ++i)
		if (((*i).mCol+1) > cols)
			cols = (*i).mCol+1;
	return cols;
}
void abstractTable::addCell(const tabCell &c)
{
#ifdef DEBUG
	cout << "addCell(row=" << c.mRow << ",col=" << c.mCol << ",rowspan=" << c.mRowspan << ",colspan=" << c.mColspan << ",cell=" << c.mCellNumber << ")" << endl;
#endif
	push_back(c);
}

void abstractTable::addCell(size_t row,size_t col,size_t rowspan,size_t colspan,int cellnumber)
{
#ifdef DEBUG
	cout << "addCell(row=" << row << ",col=" << col << ",rowspan=" << rowspan << ",colspan=" << colspan << ",cell=" << cellnumber << ")" << endl;
#endif
	push_back(tabCell(row,col,rowspan,colspan,cellnumber));
}

bool	abstractTable::existsCellAt(size_t row,size_t col) const
{
	for (const_iterator i = begin(); i != end(); ++i)
		{
		if (((*i).mRow <= row) && ((*i).mRow + (*i).mRowspan > row)
			&& ((*i).mCol <= col) && ((*i).mCol + (*i).mColspan > col))
			return true;
		}
	return false;
}

bool	abstractTable::existsCellExactlyAt(size_t row,size_t col) const
{
	for (const_iterator i = begin(); i != end(); ++i)
		{
		if (((*i).mRow == row) && ((*i).mCol == col))
			return true;
		}
	return false;
}

const tabCell&	abstractTable::findCellAt(size_t row,size_t col) const
{
	for (const_iterator i = begin(); i != end(); ++i)
		{
		if (((*i).mRow <= row) && ((*i).mRow + (*i).mRowspan > row) && ((*i).mCol <= col) && ((*i).mCol + (*i).mColspan > col))
			return (*i);
		}
	throw runtime_error("cell not found");
}

const tabCell&	abstractTable::findCellExactlyAt(size_t row,size_t col) const
{
	for (const_iterator i = begin(); i != end(); ++i)
		{
		if (((*i).mRow == row) && ((*i).mCol == col))
			return (*i);
		}
	throw runtime_error("cell not found");
}

bool abstractTable::removeCellAt(size_t row,size_t col)
{
	for (iterator i = begin(); i != end(); ++i)
		{
		if (((*i).mRow <= row) && ((*i).mRow + (*i).mRowspan > row)
			&& ((*i).mCol <= col) && ((*i).mCol + (*i).mColspan > col))
			{
			erase(i);
#ifdef DEBUG
			cout << "removeCell(row=" << row << ",col=" << col << ")" << endl;
#endif
			return true;
			}
		}

	return false;
}

void abstractTable::vbox(size_t row,size_t col,size_t rowspan)
{
	size_t hsize = 0;
	size_t vtotal = 0;

	bool done = false;

	while (!done)
		{
		vtotal = 0;
		hsize = 0;
		// find max width for this column
		for (size_t y = row; y < (row + rowspan); )
			{
			const tabCell &c = findCellAt(y,col);
			if (c.mColspan > hsize)
				hsize = c.mColspan;
			vtotal += c.mRowspan;
			y += c.mRowspan;
			}

		// build hboxes for all that are not wide enough
		done = true;
		for (size_t y = row; y < (row + rowspan); )
			{
			const tabCell &c = findCellAt(y,col);
			if (c.mColspan < hsize)
				{
				hbox(y,col,hsize);
				done = false;
				break;
				}
			y += c.mRowspan;
			}
		}

	const tabCell &c = findCellAt(row,col);
	int new_cell_number = c.mCellNumber;
	size_t new_colspan = c.mColspan;

	cout << "\\sbox{\\box" << val_to_suffix(new_cell_number) << "}{\\vbox{";

	for (size_t y = row; y < (row + rowspan); y++)
		{
		try {
			const tabCell &cell = findCellExactlyAt(y,col);
			if (y != row)
				cout << "\\nointerlineskip";
			cout << "\\hbox{\\usebox\\box" << val_to_suffix(cell.mCellNumber) << "}";
			}
		catch(...){}
		removeCellAt(y,col);
		}

	cout << "}}" << endl;

	addCell(row,col,vtotal,new_colspan,new_cell_number);
#ifdef DEBUG
	cout << "vbox(row=" << row << ",col=" << col << ",rowspan=" << rowspan << ")" << endl;
	cout << (*this);
#endif
}

void abstractTable::hbox(size_t row,size_t col,size_t colspan)
{
	size_t vsize = 0;
	size_t htotal = 0;

	bool done = false;
	while (!done)
		{
		htotal = 0;
		vsize = 0;
		// find max height for this row
		for (size_t x = col; x < (col + colspan); )
			{
			const tabCell &c = findCellAt(row,x);
			if (c.mRowspan > vsize)
				vsize = c.mRowspan;
			htotal += c.mColspan;
			x += c.mColspan;
			}

		// build vboxes for all that are not high enough
		done = true;
		for (size_t x = col; x < (col + colspan); )
			{
			const tabCell &c = findCellAt(row,x);
			if (c.mRowspan < vsize)
				{
				vbox(row,x,vsize);
				done = false;
				break;
				}
			x += c.mColspan;
			}
		}

	const tabCell &c = findCellAt(row,col);
	int new_cell_number = c.mCellNumber;
	size_t new_rowspan = c.mRowspan;

	cout << "\\sbox{\\box" << val_to_suffix(new_cell_number) << "}{";

	for (size_t x = col; x < (col + colspan); x++)
		{
		try {
			const tabCell &cell = findCellExactlyAt(row,x);
			cout << "\\usebox\\box" << val_to_suffix(cell.mCellNumber);
			}
		catch(...){} // this is ok
		removeCellAt(row,x);
		}
	cout << "}" << endl;

	addCell(row,col,new_rowspan,htotal,new_cell_number);
#ifdef DEBUG
	cout << "hbox(row=" << row << ",col=" << col << ",colspan=" << colspan << ")" << endl;
	cout << (*this);
#endif
}

void abstractTable::box()
{
	for (const_iterator i = begin(); i != end(); ++i)
		{
		cout << "\\nmrcolrebox{\\box" << val_to_suffix((*i).mCellNumber) << "}{";
		for (size_t x = 0; x < (*i).mRowspan; ++x)
			{
			if (x)
				cout << " + ";
			cout << "\\rowheight" << val_to_suffix((*i).mRow+x);
			}
		if ((*i).mRowspan > 1)
			{
			cout << " + " << (2 * ((*i).mRowspan - 1)) << "\\fboxsep + " << (2 * ((*i).mRowspan - 1)) << "\\fboxrule";
			}
		cout << "}{black}{\\ccbox" << val_to_suffix((*i).mCellNumber) << "}{\\calign" << val_to_suffix((*i).mCellNumber) << "}" << endl;
		}	

	size_t cols = getCols();

	while (size())
		{
#ifdef DEBUG
		cout << (*this);
#endif
		hbox(0,0,cols);
		const tabCell &c = findCellAt(0,0);
		size_t rowspan = c.mRowspan;
		removeCellAt(0,0);
		for(iterator i = begin(); i != end(); ++i)
			(*i).mRow -= rowspan;
		}
	cout << "{\\usebox\\boxa}\\par\\nointerlineskip" << endl;
}


ostream& operator<<(ostream &o,const abstractTable &t)
{
#if 0
	for (abstractTable::const_iterator i = t.begin(); i != t.end(); ++i)
		{
		o << (*i) << endl;
		}
#endif
	size_t rows = t.getRows();
	size_t cols = t.getCols();

	for (size_t y = 0; y < rows; ++y)
		{
		safe_ostrstream topBorder;
		safe_ostrstream line;
		string label;

		for (size_t x = 0; x < cols; ++x)
			{
			string borderstring = "+-------";
			const tabCell &cell = t.findCellAt(y,x);
			char leftLineDelim = '|';
			int cellnumber = cell.mCellNumber;
			char buf[255];
			sprintf(buf,"%d",cellnumber);
			label = buf;

			if (y)
				{
				const tabCell &topcell = t.findCellAt(y-1,x);
				if (topcell.mCellNumber == cell.mCellNumber)
					{
					label = " ";
					borderstring = "+       ";
					}
				}
			topBorder << borderstring;
			if (x)
				{
				const tabCell &leftcell = t.findCellAt(y,x-1);
				if (leftcell.mCellNumber == cell.mCellNumber)
					{
					leftLineDelim = ' ';
					label = " ";
					}
				}
			line << leftLineDelim << label << "\t";
			if (x == cols-1)
				{
				line << "|";
				topBorder << "+";
				}
			}
		o << topBorder.str() << endl;
		o << line.str() << endl;
		}
	for (size_t x = 0; x < cols; ++x)
		o << "+-------";
	o << "+" << endl;
	return o;
}

// the same as an abstract table but with different semantics. The column numbers of the
// data have not been offset for rowspan elements.
void HTMLabstractTable::box() { assert(0); throw runtime_error("Do not call HTMLabstractTable::box()");}
abstractTable HTMLabstractTable::makeAbstract() const
{
	abstractTable newtable;

	size_t cols = getCols();
	size_t rows = getRows();
	vector<int>	spans;
	for (size_t x = 0; x < cols; ++x)
		spans.push_back(0);

	for (size_t y = 0; y < rows; ++y)
		{
		size_t offset = 0;
		for (size_t x = 0; x < cols; ++x)
			{
			if (spans[x])
				{
				offset++;
				spans[x]--;
				}
			else if (existsCellExactlyAt(y,x - offset))
				{
				tabCell c = findCellExactlyAt(y,x-offset);
				c.mCol = x;
				if (c.mRowspan >= 1) // paranoia
					spans[x] = c.mRowspan - 1;
				newtable.addCell(c);
				}
			}
		}

	return newtable;
}

