#include <sys/param.h>
#include <sys/errno.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "nmr_fetch.h"
#include "common.h"


/*** Local data **************************************************************/

/*
 * Error messages for parser errors
 */
#define URL_MALFORMED		1
#define URL_BAD_SCHEME		2
#define URL_BAD_PORT		3
static struct fetcherr _url_errlist[] = {
    { URL_MALFORMED,	FETCH_URL,	"Malformed URL" },
    { URL_BAD_SCHEME,	FETCH_URL,	"Invalid URL scheme" },
    { URL_BAD_PORT,	FETCH_URL,	"Invalid server port" },
    { -1,		FETCH_UNKNOWN,	"Unknown parser error" }
};

/*
 * Split an URL into components. URL syntax is:
 * [method:/][/[user[:pwd]@]host[:port]/][document]
 * This almost, but not quite, RFC1738 URL syntax.
 */
struct url *
fetchParseURL(char *URL)
{
    char *doc, *p, *q;
    struct url *u;
    int i;

    /* allocate struct url */
    if ((u = calloc(1, sizeof *u)) == NULL) {
	_fetch_syserr();
	return NULL;
    }

    /* scheme name */
    if ((p = strstr(URL, ":/"))) {
	snprintf(u->scheme, URL_SCHEMELEN+1, "%.*s", p - URL, URL);
	URL = ++p;
	/*
	 * Only one slash: no host, leave slash as part of document
	 * Two slashes: host follows, strip slashes
	 */
	if (URL[1] == '/')
	    URL = (p += 2);
    } else {
	p = URL;
    }
    if (!*URL || *URL == '/')
	goto nohost;

    p = strpbrk(URL, "/@");
    if (p && *p == '@') {
	/* username */
	for (q = URL, i = 0; (*q != ':') && (*q != '@'); q++)
	    if (i < URL_USERLEN)
		u->user[i++] = *q;
	
	/* password */
	if (*q == ':')
	    for (q++, i = 0; (*q != ':') && (*q != '@'); q++)
		if (i < URL_PWDLEN)
		    u->pwd[i++] = *q;
	
	p++;
    } else p = URL;
    
    /* hostname */
#ifdef INET6
    if (*p == '[' && (q = strchr(p + 1, ']')) != NULL &&
	(*++q == '\0' || *q == '/' || *q == ':')) {
	if ((i = q - p - 2) > MAXHOSTNAMELEN)
	    i = MAXHOSTNAMELEN;
	strncpy(u->host, ++p, i);
	p = q;
    } else
#endif
	for (i = 0; *p && (*p != '/') && (*p != ':'); p++)
	    if (i < MAXHOSTNAMELEN)
		u->host[i++] = *p;

    /* port */
    if (*p == ':') {
	for (q = ++p; *q && (*q != '/'); q++)
	    if (isdigit(*q))
		u->port = u->port * 10 + (*q - '0');
	    else {
		/* invalid port */
		_url_seterr(URL_BAD_PORT);
		goto ouch;
	    }
	while (*p && (*p != '/'))
	    p++;
    }

nohost:
    /* document */
    if (!*p)
	p = "/";
    
    if (strcasecmp(u->scheme, SCHEME_HTTP) == 0 ||
	strcasecmp(u->scheme, SCHEME_HTTPS) == 0) {
	const char hexnums[] = "0123456789abcdef";

	/* percent-escape whitespace. */
	if ((doc = malloc(strlen(p) * 3 + 1)) == NULL) {
	    _fetch_syserr();
	    goto ouch;
	}
	u->doc = doc;
	while (*p != '\0') {
	    if (!isspace(*p)) {
		*doc++ = *p++;
            } else {
		*doc++ = '%';
		*doc++ = hexnums[((unsigned int)*p) >> 4];
		*doc++ = hexnums[((unsigned int)*p) & 0xf];
		p++;
            }
	}
	*doc = '\0';
    } else if ((u->doc = strdup(p)) == NULL) {
	_fetch_syserr();
	goto ouch;
    }
    
    DEBUG(fprintf(stderr,
		  "scheme:   [\033[1m%s\033[m]\n"
		  "user:     [\033[1m%s\033[m]\n"
		  "password: [\033[1m%s\033[m]\n"
		  "host:     [\033[1m%s\033[m]\n"
		  "port:     [\033[1m%d\033[m]\n"
		  "document: [\033[1m%s\033[m]\n",
		  u->scheme, u->user, u->pwd,
		  u->host, u->port, u->doc));
    
    return u;

ouch:
    free(u);
    return NULL;
}

/*
 * Free a URL
 */
void
fetchFreeURL(struct url *u)
{
    free(u->doc);
    free(u);
}
