/*
 * This file is part of htmltolatex (h2l) a HTML to Latex converter
 * Copyright (C) 2001 Neil Radisch (nradisch@panix.com)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file copying.txt.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */
#include "util.h"
#include <map>
#include <ctype.h>
#include <sys/param.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include "safe_ostrstream.h"
#ifdef LINUX
extern "C" {
#include "myfetch/nmr_fetch.h"
}
#else
extern "C"{
#include "fetch.h"
}
#endif
using namespace std;

static map<string,int> gsSpecialTable;

class init_special 
{ 
public: init_special()
	{
	gsSpecialTable["quot"]		= 34;
	gsSpecialTable["amp"] 		= 38;
	gsSpecialTable["lt"] 		= 60;
	gsSpecialTable["gt"] 		= 62;
	gsSpecialTable["nbsp"] 		= 160;
	gsSpecialTable["iexcl"] 	= 161;
	gsSpecialTable["cent"] 		= 162;
	gsSpecialTable["pound"] 	= 163;
	gsSpecialTable["curren"] 	= 164;
	gsSpecialTable["yen"] 		= 165;
	gsSpecialTable["brvbar"] 	= 166;
	gsSpecialTable["sect"] 		= 167;
	gsSpecialTable["uml"] 		= 168;
	gsSpecialTable["copy"] 		= 169;
	gsSpecialTable["ordf"] 		= 170;
	gsSpecialTable["laquo"] 	= 171;
	gsSpecialTable["not"] 		= 172;
	gsSpecialTable["shy"] 		= 173;
	gsSpecialTable["reg"] 		= 174;
	gsSpecialTable["macr"] 		= 175;
	gsSpecialTable["deg"] 		= 176;
	gsSpecialTable["plusmn"] 	= 177; 
	gsSpecialTable["sup2"] 		= 178;
	gsSpecialTable["sup3"] 		= 179;
	gsSpecialTable["acute"] 	= 180;
	gsSpecialTable["micro"] 	= 181;
	gsSpecialTable["para"] 		= 182;
	gsSpecialTable["middot"] 	= 183; 
	gsSpecialTable["cedil"] 	= 184; 
	gsSpecialTable["sup1"] 		= 185; 
	gsSpecialTable["ordm"] 		= 186;
	gsSpecialTable["rquo"] 		= 187;
	gsSpecialTable["frac14"] 	= 188;
	gsSpecialTable["frac12"] 	= 189;
	gsSpecialTable["frac34"] 	= 190;
	gsSpecialTable["iquest"] 	= 191;
	gsSpecialTable["Agrave"] 	= 192;
	gsSpecialTable["Aacute"] 	= 193;
	gsSpecialTable["Acirc"] 	= 194;
	gsSpecialTable["Atilde"] 	= 195;
	gsSpecialTable["Auml"] 		= 196;
	gsSpecialTable["Aring"] 	= 197;
	gsSpecialTable["AElig"] 	= 198;
	gsSpecialTable["Ccedil"] 	= 199;
	gsSpecialTable["Egrave"] 	= 200;
	gsSpecialTable["Eacute"] 	= 201; 
	gsSpecialTable["Ecirc"] 	= 202;
	gsSpecialTable["Euml"] 		= 203;
	gsSpecialTable["Igrave"] 	= 204;
	gsSpecialTable["Iacute"] 	= 205;
	gsSpecialTable["Icric"] 	= 206;
	gsSpecialTable["Iuml"] 		= 207;
	gsSpecialTable["ETH"] 		= 208;
	gsSpecialTable["Ntilde"] 	= 209; 
	gsSpecialTable["Ograve"] 	= 210;
	gsSpecialTable["Oacute"] 	= 211;
	gsSpecialTable["Ocirc"] 	= 212;
	gsSpecialTable["Otilde"] 	= 213;
	gsSpecialTable["Ouml"] 		= 214;
	gsSpecialTable["times"] 	= 215;
	gsSpecialTable["Oslash"] 	= 216;
	gsSpecialTable["Ugrave"] 	= 217;
	gsSpecialTable["Uacute"] 	= 218;
	gsSpecialTable["Ucirc"] 	= 219;
	gsSpecialTable["Uuml"] 		= 220;
	gsSpecialTable["Yacute"] 	= 221;
	gsSpecialTable["THORN"] 	= 222;
	gsSpecialTable["szlig"] 	= 223;
	gsSpecialTable["agrave"] 	= 224;
	gsSpecialTable["aacute"] 	= 225;
	gsSpecialTable["acirc"] 	= 226;
	gsSpecialTable["atilde"] 	= 227;
	gsSpecialTable["auml"] 		= 228;
	gsSpecialTable["aring"] 	= 229;
	gsSpecialTable["aelig"] 	= 230;
	gsSpecialTable["ccedil"] 	= 231;
	gsSpecialTable["egrave"] 	= 232;
	gsSpecialTable["eacute"] 	= 233;
	gsSpecialTable["ecirc"] 	= 234;
	gsSpecialTable["euml"] 		= 235;
	gsSpecialTable["igrave"] 	= 236;
	gsSpecialTable["iacute"] 	= 237;
	gsSpecialTable["icirc"] 	= 238;
	gsSpecialTable["iuml"] 		= 239; 
	gsSpecialTable["eth"] 		= 240;
	gsSpecialTable["ntilde"] 	= 241; 
	gsSpecialTable["ograve"] 	= 242;
	gsSpecialTable["oacute"] 	= 243;
	gsSpecialTable["ocirc"] 	= 244;
	gsSpecialTable["otilde"] 	= 245;
	gsSpecialTable["ouml"] 		= 246;
	gsSpecialTable["divide"] 	= 247; 
	gsSpecialTable["oslash"] 	= 248;
	gsSpecialTable["ugrave"] 	= 249;
	gsSpecialTable["uacute"] 	= 250;
	gsSpecialTable["ucirc"] 	= 251;
	gsSpecialTable["uuml"] 		= 252;
	gsSpecialTable["yacute"] 	= 253;
	gsSpecialTable["thorn"] 	= 254;
	gsSpecialTable["yuml"] 		= 255;
	}
};

static init_special gsIS;

char lookup_special(const string &s)
{
	map<string,int>::const_iterator i = gsSpecialTable.find(s);
	if (i != gsSpecialTable.end())
		return (char)((*i).second);
	else
		return ' ';
}

string lowerstring(const string &s)
{
	string result;

	string::const_iterator i = s.begin();
	while (i != s.end())
		{
		result += tolower(*i);
		i++;
		}

	return result;
}

string stripstring(const string &s)
{
	string	accum,result;
	bool gotchar = false;

	string::const_iterator i = s.begin();
	while (i != s.end())
		{
		if (isspace((*i)))
			{
			accum += (*i);
			}
		else
			{
			if (gotchar)
				result += accum;

			accum = "";
			result += (*i);
			gotchar = true;
			}
		i++;
		}

	return result;
}

	string _remove_trailing_clause(char c,const string &s)
	{
	string result;

	string::const_iterator i = s.begin();
	while (i != s.end())
		{
		if ((*i) == c) break;
		result += (*i);
		i++;
		}

	return result;
	}

string removeequalclause(const string &s) { return _remove_trailing_clause('=',s); }
string removecolonclause(const string &s) { return _remove_trailing_clause(':',s); }

	int hexdigit(char c)
		{
		if (c >= '0' && c <= '9')
			return (c-'0');
		if (c >= 'a' && c <= 'f')
			return (c - 'a' + 10);
		if (c >= 'A' && c <= 'F')
			return(c - 'A' + 10);

		return 0;
		}

void htmlrgbtotexrgb(const std::string &s,double &r,double &g,double &b)
{
	// takes string of form "#xxxxxx" only
	r = g = b = 0;

	string::const_iterator i = s.begin();
	if (i == s.end()) return;
	i++;
	if (i == s.end()) return;

	r = (hexdigit(*(i++)) * 16);	
	if (i == s.end()) { r /= 255.0; return; }
	r += hexdigit(*(i++));
	r /= 255.0;
	if (i == s.end()) return;

	g = (hexdigit(*(i++)) * 16);	
	if (i == s.end()) { g /= 255.0; return; }
	g += hexdigit(*(i++));
	g /= 255.0;
	if (i == s.end()) return;
	
	b = (hexdigit(*(i++)) * 16);	
	if (i == s.end()) { b /= 255.0; return; }
	b += hexdigit(*(i++));
	b /= 255.0;

#if 0
	r = (hexdigit(s[1]) * 16) + hexdigit(s[2]);	
	g = (hexdigit(s[3]) * 16) + hexdigit(s[4]);	
	b = (hexdigit(s[5]) * 16) + hexdigit(s[6]);	
	r /= 255.0;
	g /= 255.0;
	b /= 255.0;
#endif
}

bool allhexdigits(const string &s)
{
	for (string::const_iterator i = s.begin(); i != s.end(); ++i)
		{
		char c = (*i);
		if (c >= '0' && c <= '9')
			continue;
		if (c >= 'a' && c <= 'f')
			continue;
		if (c >= 'A' && c <= 'F')
			continue;
		return false;
		}
	return true;
}


// path begins with slash if server absolute, else current doc relative
// path will always end in a slash unless empty
void split_url(const string &url,string &server,string &path,string &file)
{
	server = "";
	path = "";
	file = "";

	struct url *u = fetchParseURL(const_cast<char*>((stripstring(url)).c_str()));
	if (!u) return;

	// now build the stuff I'm interested in
	if (strlen(u->scheme))
		{
		server += string(u->scheme) + "://";

		if (strlen(u->user))
			{
			server += u->user;
			if (strlen(u->pwd))
				server += string(":") + u->pwd;
			server += "@";
			}

		if (strlen(u->host))
			server += u->host;

		if (u->port != 80 && u->port != 0)
			{
			safe_ostrstream o;
			o << u->port;
			server += string(":") + o.str();
			}
		path = u->doc;
		}
	else
		{
		path = string(u->host);
		size_t len = strlen(u->doc);
		if (len)
			{
			if (len > 1 || (u->doc)[0] != '/')
				path += u->doc;
			}
		}

	// now the path must manually be split
	// scan forward until question mark or end of string
	string::size_type last_slash = string::npos;
	string::size_type last_dot = string::npos;

	for (string::size_type i = 0; i < path.size(); ++i)
		{
		if (path[i] == '?') break;

		if (path[i] == '/') 
			last_slash = i;
		else if (path[i] == '.')
			last_dot = i;
		}

	if ((last_slash == string::npos) && (last_dot != string::npos))
		{
		file = path;
		path = "";
		}
	else if (last_dot == string::npos)
		{
		file = "index.html";
		}
	else if (last_dot < last_slash)
		{
		file = "index.html";
		}
	else
		{
		string t = path;
		path = t.substr(0,last_slash+1);
		file = t.substr(last_slash+1,t.size() - last_slash);
		}

	if (path != "" && path[path.size() - 1] != '/')
		path += '/';


	fetchFreeURL(u); u = 0;

	server = lowerstring(server); // only server can be lowercased
}


bool can_spider_url(const string &url)
{
	string s = lowerstring(stripstring(url));
	if (s == "") return false;

	if (s.find_last_of(".html") == url.size() - 1)
		return true;
	if (s.find_last_of(".php") == url.size() - 1)
		return true;
	if (s.find_last_of(".jsp") == url.size() - 1)
		return true;
	if (s.find_last_of(".asp") == url.size() - 1)
		return true;

	return false;
}

string cleanurl(const string &s)
{
	string result;
	for (string::const_iterator i = s.begin(); i != s.end(); ++i)
		{
		if ((*i) == '}')
			result += "\\}";
		else if ((*i) == '{')
			result += "\\{";
#if 0
		else if ((*i) == '_')
			result += "\\_";
#endif
		else if ((*i) == '&')
			result += "\\&";
		else if ((*i) == '$')
			result += "\\$";
		else if ((*i) == '%')
			result += "\\%";
		else if ((*i) == '#')
			result += "\\#";
		else
			result += (*i);
		}
	return result;
}

	string val_to_suffix(size_t v)
	{
	string result;
	result = result + (char)('a' + (v % 26));
	if (v > 25)
		result = val_to_suffix(v / 26) + result;
	return result;
	}

